# -* coding: utf-8 -*
# Chemfiles, a modern library for chemistry file reading and writing
# Copyright (C) Guillaume Fraux and contributors -- BSD license
#
# =========================================================================== #
# !!!! AUTO-GENERATED FILE !!!! Do not edit. See the bindgen repository for
# the generation code (https://github.com/chemfiles/bindgen).
# This file contains Python ctype interface to the C API
# =========================================================================== #

# flake8: noqa
'''
Foreign function interface declaration for the Python interface to chemfiles
'''
from numpy.ctypeslib import ndpointer
import numpy as np
from ctypes import c_int, c_uint64, c_double, c_char, c_char_p, c_void_p, c_bool
from ctypes import CFUNCTYPE, ARRAY, POINTER, Structure

from ._utils import _check_return_code


class chfl_status(c_int):
    CHFL_SUCCESS = 0
    CHFL_MEMORY_ERROR = 1
    CHFL_FILE_ERROR = 2
    CHFL_FORMAT_ERROR = 3
    CHFL_SELECTION_ERROR = 4
    CHFL_CONFIGURATION_ERROR = 5
    CHFL_OUT_OF_BOUNDS = 6
    CHFL_PROPERTY_ERROR = 7
    CHFL_GENERIC_ERROR = 254
    CHFL_CXX_ERROR = 255


class chfl_bond_order(c_int):
    CHFL_BOND_UNKNOWN = 0
    CHFL_BOND_SINGLE = 1
    CHFL_BOND_DOUBLE = 2
    CHFL_BOND_TRIPLE = 3
    CHFL_BOND_QUADRUPLE = 4
    CHFL_BOND_QINTUPLET = 5
    CHFL_BOND_AMIDE = 254
    CHFL_BOND_AROMATIC = 255


class chfl_property_kind(c_int):
    CHFL_PROPERTY_BOOL = 0
    CHFL_PROPERTY_DOUBLE = 1
    CHFL_PROPERTY_STRING = 2
    CHFL_PROPERTY_VECTOR3D = 3


class chfl_cellshape(c_int):
    CHFL_CELL_ORTHORHOMBIC = 0
    CHFL_CELL_TRICLINIC = 1
    CHFL_CELL_INFINITE = 2


class CHFL_TRAJECTORY(Structure):
    pass


class CHFL_CELL(Structure):
    pass


class CHFL_ATOM(Structure):
    pass


class CHFL_FRAME(Structure):
    pass


class CHFL_TOPOLOGY(Structure):
    pass


class CHFL_SELECTION(Structure):
    pass


class CHFL_RESIDUE(Structure):
    pass


class CHFL_PROPERTY(Structure):
    pass

# Some hand-defined type. Make sure to edit the bindgen code to make this
# correspond to the current chemfiles.h header
chfl_vector3d = ARRAY(c_double, 3)

chfl_warning_callback = CFUNCTYPE(None, c_char_p)

class chfl_match(Structure):
    _fields_ = [
        ('size', c_uint64),
        ('atoms', ARRAY(c_uint64, 4))
    ]

# end of hand-defined types


def set_interface(c_lib):
    from chemfiles import Atom
    from chemfiles import Residue
    from chemfiles import Topology
    from chemfiles import UnitCell
    from chemfiles import Frame
    from chemfiles import Selection
    from chemfiles import Trajectory

    from chemfiles import Property

    # Manually defined functions
    c_lib.chfl_free.argtypes = [c_void_p]
    c_lib.chfl_trajectory_close.argtypes = [Trajectory]
    # End of manually defined functions

    # Function "chfl_version", at types.h:145
    c_lib.chfl_version.argtypes = []
    c_lib.chfl_version.restype = c_char_p

    # Function "chfl_last_error", at misc.h:19
    c_lib.chfl_last_error.argtypes = []
    c_lib.chfl_last_error.restype = c_char_p

    # Function "chfl_clear_errors", at misc.h:29
    c_lib.chfl_clear_errors.argtypes = []
    c_lib.chfl_clear_errors.restype = chfl_status
    c_lib.chfl_clear_errors.errcheck = _check_return_code

    # Function "chfl_set_warning_callback", at misc.h:38
    c_lib.chfl_set_warning_callback.argtypes = [chfl_warning_callback]
    c_lib.chfl_set_warning_callback.restype = chfl_status
    c_lib.chfl_set_warning_callback.errcheck = _check_return_code

    # Function "chfl_add_configuration", at misc.h:54
    c_lib.chfl_add_configuration.argtypes = [c_char_p]
    c_lib.chfl_add_configuration.restype = chfl_status
    c_lib.chfl_add_configuration.errcheck = _check_return_code

    # Function "chfl_property_bool", at property.h:32
    c_lib.chfl_property_bool.argtypes = [c_bool]
    c_lib.chfl_property_bool.restype = POINTER(CHFL_PROPERTY)

    # Function "chfl_property_double", at property.h:42
    c_lib.chfl_property_double.argtypes = [c_double]
    c_lib.chfl_property_double.restype = POINTER(CHFL_PROPERTY)

    # Function "chfl_property_string", at property.h:52
    c_lib.chfl_property_string.argtypes = [c_char_p]
    c_lib.chfl_property_string.restype = POINTER(CHFL_PROPERTY)

    # Function "chfl_property_vector3d", at property.h:62
    c_lib.chfl_property_vector3d.argtypes = [chfl_vector3d]
    c_lib.chfl_property_vector3d.restype = POINTER(CHFL_PROPERTY)

    # Function "chfl_property_get_kind", at property.h:69
    c_lib.chfl_property_get_kind.argtypes = [Property, POINTER(chfl_property_kind)]
    c_lib.chfl_property_get_kind.restype = chfl_status
    c_lib.chfl_property_get_kind.errcheck = _check_return_code

    # Function "chfl_property_get_bool", at property.h:82
    c_lib.chfl_property_get_bool.argtypes = [Property, POINTER(c_bool)]
    c_lib.chfl_property_get_bool.restype = chfl_status
    c_lib.chfl_property_get_bool.errcheck = _check_return_code

    # Function "chfl_property_get_double", at property.h:95
    c_lib.chfl_property_get_double.argtypes = [Property, POINTER(c_double)]
    c_lib.chfl_property_get_double.restype = chfl_status
    c_lib.chfl_property_get_double.errcheck = _check_return_code

    # Function "chfl_property_get_string", at property.h:110
    c_lib.chfl_property_get_string.argtypes = [Property, c_char_p, c_uint64]
    c_lib.chfl_property_get_string.restype = chfl_status
    c_lib.chfl_property_get_string.errcheck = _check_return_code

    # Function "chfl_property_get_vector3d", at property.h:123
    c_lib.chfl_property_get_vector3d.argtypes = [Property, chfl_vector3d]
    c_lib.chfl_property_get_vector3d.restype = chfl_status
    c_lib.chfl_property_get_vector3d.errcheck = _check_return_code

    # Function "chfl_atom", at atom.h:20
    c_lib.chfl_atom.argtypes = [c_char_p]
    c_lib.chfl_atom.restype = POINTER(CHFL_ATOM)

    # Function "chfl_atom_copy", at atom.h:30
    c_lib.chfl_atom_copy.argtypes = [Atom]
    c_lib.chfl_atom_copy.restype = POINTER(CHFL_ATOM)

    # Function "chfl_atom_from_frame", at atom.h:54
    c_lib.chfl_atom_from_frame.argtypes = [Frame, c_uint64]
    c_lib.chfl_atom_from_frame.restype = POINTER(CHFL_ATOM)

    # Function "chfl_atom_from_topology", at atom.h:77
    c_lib.chfl_atom_from_topology.argtypes = [Topology, c_uint64]
    c_lib.chfl_atom_from_topology.restype = POINTER(CHFL_ATOM)

    # Function "chfl_atom_mass", at atom.h:88
    c_lib.chfl_atom_mass.argtypes = [Atom, POINTER(c_double)]
    c_lib.chfl_atom_mass.restype = chfl_status
    c_lib.chfl_atom_mass.errcheck = _check_return_code

    # Function "chfl_atom_set_mass", at atom.h:97
    c_lib.chfl_atom_set_mass.argtypes = [Atom, c_double]
    c_lib.chfl_atom_set_mass.restype = chfl_status
    c_lib.chfl_atom_set_mass.errcheck = _check_return_code

    # Function "chfl_atom_charge", at atom.h:106
    c_lib.chfl_atom_charge.argtypes = [Atom, POINTER(c_double)]
    c_lib.chfl_atom_charge.restype = chfl_status
    c_lib.chfl_atom_charge.errcheck = _check_return_code

    # Function "chfl_atom_set_charge", at atom.h:115
    c_lib.chfl_atom_set_charge.argtypes = [Atom, c_double]
    c_lib.chfl_atom_set_charge.restype = chfl_status
    c_lib.chfl_atom_set_charge.errcheck = _check_return_code

    # Function "chfl_atom_type", at atom.h:125
    c_lib.chfl_atom_type.argtypes = [Atom, c_char_p, c_uint64]
    c_lib.chfl_atom_type.restype = chfl_status
    c_lib.chfl_atom_type.errcheck = _check_return_code

    # Function "chfl_atom_set_type", at atom.h:136
    c_lib.chfl_atom_set_type.argtypes = [Atom, c_char_p]
    c_lib.chfl_atom_set_type.restype = chfl_status
    c_lib.chfl_atom_set_type.errcheck = _check_return_code

    # Function "chfl_atom_name", at atom.h:146
    c_lib.chfl_atom_name.argtypes = [Atom, c_char_p, c_uint64]
    c_lib.chfl_atom_name.restype = chfl_status
    c_lib.chfl_atom_name.errcheck = _check_return_code

    # Function "chfl_atom_set_name", at atom.h:157
    c_lib.chfl_atom_set_name.argtypes = [Atom, c_char_p]
    c_lib.chfl_atom_set_name.restype = chfl_status
    c_lib.chfl_atom_set_name.errcheck = _check_return_code

    # Function "chfl_atom_full_name", at atom.h:167
    c_lib.chfl_atom_full_name.argtypes = [Atom, c_char_p, c_uint64]
    c_lib.chfl_atom_full_name.restype = chfl_status
    c_lib.chfl_atom_full_name.errcheck = _check_return_code

    # Function "chfl_atom_vdw_radius", at atom.h:179
    c_lib.chfl_atom_vdw_radius.argtypes = [Atom, POINTER(c_double)]
    c_lib.chfl_atom_vdw_radius.restype = chfl_status
    c_lib.chfl_atom_vdw_radius.errcheck = _check_return_code

    # Function "chfl_atom_covalent_radius", at atom.h:189
    c_lib.chfl_atom_covalent_radius.argtypes = [Atom, POINTER(c_double)]
    c_lib.chfl_atom_covalent_radius.restype = chfl_status
    c_lib.chfl_atom_covalent_radius.errcheck = _check_return_code

    # Function "chfl_atom_atomic_number", at atom.h:199
    c_lib.chfl_atom_atomic_number.argtypes = [Atom, POINTER(c_uint64)]
    c_lib.chfl_atom_atomic_number.restype = chfl_status
    c_lib.chfl_atom_atomic_number.errcheck = _check_return_code

    # Function "chfl_atom_properties_count", at atom.h:206
    c_lib.chfl_atom_properties_count.argtypes = [Atom, POINTER(c_uint64)]
    c_lib.chfl_atom_properties_count.restype = chfl_status
    c_lib.chfl_atom_properties_count.errcheck = _check_return_code

    # Function "chfl_atom_list_properties", at atom.h:222
    c_lib.chfl_atom_list_properties.argtypes = [Atom, POINTER(c_char_p), c_uint64]
    c_lib.chfl_atom_list_properties.restype = chfl_status
    c_lib.chfl_atom_list_properties.errcheck = _check_return_code

    # Function "chfl_atom_set_property", at atom.h:234
    c_lib.chfl_atom_set_property.argtypes = [Atom, c_char_p, Property]
    c_lib.chfl_atom_set_property.restype = chfl_status
    c_lib.chfl_atom_set_property.errcheck = _check_return_code

    # Function "chfl_atom_get_property", at atom.h:248
    c_lib.chfl_atom_get_property.argtypes = [Atom, c_char_p]
    c_lib.chfl_atom_get_property.restype = POINTER(CHFL_PROPERTY)

    # Function "chfl_residue", at residue.h:20
    c_lib.chfl_residue.argtypes = [c_char_p]
    c_lib.chfl_residue.restype = POINTER(CHFL_RESIDUE)

    # Function "chfl_residue_with_id", at residue.h:30
    c_lib.chfl_residue_with_id.argtypes = [c_char_p, c_uint64]
    c_lib.chfl_residue_with_id.restype = POINTER(CHFL_RESIDUE)

    # Function "chfl_residue_from_topology", at residue.h:52
    c_lib.chfl_residue_from_topology.argtypes = [Topology, c_uint64]
    c_lib.chfl_residue_from_topology.restype = POINTER(CHFL_RESIDUE)

    # Function "chfl_residue_for_atom", at residue.h:73
    c_lib.chfl_residue_for_atom.argtypes = [Topology, c_uint64]
    c_lib.chfl_residue_for_atom.restype = POINTER(CHFL_RESIDUE)

    # Function "chfl_residue_copy", at residue.h:85
    c_lib.chfl_residue_copy.argtypes = [Residue]
    c_lib.chfl_residue_copy.restype = POINTER(CHFL_RESIDUE)

    # Function "chfl_residue_atoms_count", at residue.h:92
    c_lib.chfl_residue_atoms_count.argtypes = [Residue, POINTER(c_uint64)]
    c_lib.chfl_residue_atoms_count.restype = chfl_status
    c_lib.chfl_residue_atoms_count.errcheck = _check_return_code

    # Function "chfl_residue_atoms", at residue.h:106
    c_lib.chfl_residue_atoms.argtypes = [Residue, ndpointer(np.uint64, flags="C_CONTIGUOUS", ndim=1), c_uint64]
    c_lib.chfl_residue_atoms.restype = chfl_status
    c_lib.chfl_residue_atoms.errcheck = _check_return_code

    # Function "chfl_residue_id", at residue.h:119
    c_lib.chfl_residue_id.argtypes = [Residue, POINTER(c_uint64)]
    c_lib.chfl_residue_id.restype = chfl_status
    c_lib.chfl_residue_id.errcheck = _check_return_code

    # Function "chfl_residue_name", at residue.h:131
    c_lib.chfl_residue_name.argtypes = [Residue, c_char_p, c_uint64]
    c_lib.chfl_residue_name.restype = chfl_status
    c_lib.chfl_residue_name.errcheck = _check_return_code

    # Function "chfl_residue_add_atom", at residue.h:140
    c_lib.chfl_residue_add_atom.argtypes = [Residue, c_uint64]
    c_lib.chfl_residue_add_atom.restype = chfl_status
    c_lib.chfl_residue_add_atom.errcheck = _check_return_code

    # Function "chfl_residue_contains", at residue.h:150
    c_lib.chfl_residue_contains.argtypes = [Residue, c_uint64, POINTER(c_bool)]
    c_lib.chfl_residue_contains.restype = chfl_status
    c_lib.chfl_residue_contains.errcheck = _check_return_code

    # Function "chfl_residue_properties_count", at residue.h:159
    c_lib.chfl_residue_properties_count.argtypes = [Residue, POINTER(c_uint64)]
    c_lib.chfl_residue_properties_count.restype = chfl_status
    c_lib.chfl_residue_properties_count.errcheck = _check_return_code

    # Function "chfl_residue_list_properties", at residue.h:175
    c_lib.chfl_residue_list_properties.argtypes = [Residue, POINTER(c_char_p), c_uint64]
    c_lib.chfl_residue_list_properties.restype = chfl_status
    c_lib.chfl_residue_list_properties.errcheck = _check_return_code

    # Function "chfl_residue_set_property", at residue.h:187
    c_lib.chfl_residue_set_property.argtypes = [Residue, c_char_p, Property]
    c_lib.chfl_residue_set_property.restype = chfl_status
    c_lib.chfl_residue_set_property.errcheck = _check_return_code

    # Function "chfl_residue_get_property", at residue.h:201
    c_lib.chfl_residue_get_property.argtypes = [Residue, c_char_p]
    c_lib.chfl_residue_get_property.restype = POINTER(CHFL_PROPERTY)

    # Function "chfl_topology", at topology.h:20
    c_lib.chfl_topology.argtypes = []
    c_lib.chfl_topology.restype = POINTER(CHFL_TOPOLOGY)

    # Function "chfl_topology_from_frame", at topology.h:34
    c_lib.chfl_topology_from_frame.argtypes = [Frame]
    c_lib.chfl_topology_from_frame.restype = POINTER(CHFL_TOPOLOGY)

    # Function "chfl_topology_copy", at topology.h:44
    c_lib.chfl_topology_copy.argtypes = [Topology]
    c_lib.chfl_topology_copy.restype = POINTER(CHFL_TOPOLOGY)

    # Function "chfl_topology_atoms_count", at topology.h:52
    c_lib.chfl_topology_atoms_count.argtypes = [Topology, POINTER(c_uint64)]
    c_lib.chfl_topology_atoms_count.restype = chfl_status
    c_lib.chfl_topology_atoms_count.errcheck = _check_return_code

    # Function "chfl_topology_resize", at topology.h:64
    c_lib.chfl_topology_resize.argtypes = [Topology, c_uint64]
    c_lib.chfl_topology_resize.restype = chfl_status
    c_lib.chfl_topology_resize.errcheck = _check_return_code

    # Function "chfl_topology_add_atom", at topology.h:71
    c_lib.chfl_topology_add_atom.argtypes = [Topology, Atom]
    c_lib.chfl_topology_add_atom.restype = chfl_status
    c_lib.chfl_topology_add_atom.errcheck = _check_return_code

    # Function "chfl_topology_remove", at topology.h:82
    c_lib.chfl_topology_remove.argtypes = [Topology, c_uint64]
    c_lib.chfl_topology_remove.restype = chfl_status
    c_lib.chfl_topology_remove.errcheck = _check_return_code

    # Function "chfl_topology_bonds_count", at topology.h:91
    c_lib.chfl_topology_bonds_count.argtypes = [Topology, POINTER(c_uint64)]
    c_lib.chfl_topology_bonds_count.restype = chfl_status
    c_lib.chfl_topology_bonds_count.errcheck = _check_return_code

    # Function "chfl_topology_angles_count", at topology.h:100
    c_lib.chfl_topology_angles_count.argtypes = [Topology, POINTER(c_uint64)]
    c_lib.chfl_topology_angles_count.restype = chfl_status
    c_lib.chfl_topology_angles_count.errcheck = _check_return_code

    # Function "chfl_topology_dihedrals_count", at topology.h:109
    c_lib.chfl_topology_dihedrals_count.argtypes = [Topology, POINTER(c_uint64)]
    c_lib.chfl_topology_dihedrals_count.restype = chfl_status
    c_lib.chfl_topology_dihedrals_count.errcheck = _check_return_code

    # Function "chfl_topology_impropers_count", at topology.h:118
    c_lib.chfl_topology_impropers_count.argtypes = [Topology, POINTER(c_uint64)]
    c_lib.chfl_topology_impropers_count.restype = chfl_status
    c_lib.chfl_topology_impropers_count.errcheck = _check_return_code

    # Function "chfl_topology_bonds", at topology.h:131
    c_lib.chfl_topology_bonds.argtypes = [Topology, ndpointer(np.uint64, flags="C_CONTIGUOUS", ndim=2), c_uint64]
    c_lib.chfl_topology_bonds.restype = chfl_status
    c_lib.chfl_topology_bonds.errcheck = _check_return_code

    # Function "chfl_topology_angles", at topology.h:144
    c_lib.chfl_topology_angles.argtypes = [Topology, ndpointer(np.uint64, flags="C_CONTIGUOUS", ndim=2), c_uint64]
    c_lib.chfl_topology_angles.restype = chfl_status
    c_lib.chfl_topology_angles.errcheck = _check_return_code

    # Function "chfl_topology_dihedrals", at topology.h:158
    c_lib.chfl_topology_dihedrals.argtypes = [Topology, ndpointer(np.uint64, flags="C_CONTIGUOUS", ndim=2), c_uint64]
    c_lib.chfl_topology_dihedrals.restype = chfl_status
    c_lib.chfl_topology_dihedrals.errcheck = _check_return_code

    # Function "chfl_topology_impropers", at topology.h:172
    c_lib.chfl_topology_impropers.argtypes = [Topology, ndpointer(np.uint64, flags="C_CONTIGUOUS", ndim=2), c_uint64]
    c_lib.chfl_topology_impropers.restype = chfl_status
    c_lib.chfl_topology_impropers.errcheck = _check_return_code

    # Function "chfl_topology_add_bond", at topology.h:181
    c_lib.chfl_topology_add_bond.argtypes = [Topology, c_uint64, c_uint64]
    c_lib.chfl_topology_add_bond.restype = chfl_status
    c_lib.chfl_topology_add_bond.errcheck = _check_return_code

    # Function "chfl_topology_remove_bond", at topology.h:193
    c_lib.chfl_topology_remove_bond.argtypes = [Topology, c_uint64, c_uint64]
    c_lib.chfl_topology_remove_bond.restype = chfl_status
    c_lib.chfl_topology_remove_bond.errcheck = _check_return_code

    # Function "chfl_topology_residues_count", at topology.h:203
    c_lib.chfl_topology_residues_count.argtypes = [Topology, POINTER(c_uint64)]
    c_lib.chfl_topology_residues_count.restype = chfl_status
    c_lib.chfl_topology_residues_count.errcheck = _check_return_code

    # Function "chfl_topology_add_residue", at topology.h:215
    c_lib.chfl_topology_add_residue.argtypes = [Topology, Residue]
    c_lib.chfl_topology_add_residue.restype = chfl_status
    c_lib.chfl_topology_add_residue.errcheck = _check_return_code

    # Function "chfl_topology_residues_linked", at topology.h:226
    c_lib.chfl_topology_residues_linked.argtypes = [Topology, Residue, Residue, POINTER(c_bool)]
    c_lib.chfl_topology_residues_linked.restype = chfl_status
    c_lib.chfl_topology_residues_linked.errcheck = _check_return_code

    # Function "chfl_topology_bond_with_order", at topology.h:239
    c_lib.chfl_topology_bond_with_order.argtypes = [Topology, c_uint64, c_uint64, chfl_bond_order]
    c_lib.chfl_topology_bond_with_order.restype = chfl_status
    c_lib.chfl_topology_bond_with_order.errcheck = _check_return_code

    # Function "chfl_topology_bond_orders", at topology.h:253
    c_lib.chfl_topology_bond_orders.argtypes = [Topology, ndpointer(chfl_bond_order, flags="C_CONTIGUOUS", ndim=1), c_uint64]
    c_lib.chfl_topology_bond_orders.restype = chfl_status
    c_lib.chfl_topology_bond_orders.errcheck = _check_return_code

    # Function "chfl_topology_bond_order", at topology.h:266
    c_lib.chfl_topology_bond_order.argtypes = [Topology, c_uint64, c_uint64, POINTER(chfl_bond_order)]
    c_lib.chfl_topology_bond_order.restype = chfl_status
    c_lib.chfl_topology_bond_order.errcheck = _check_return_code

    # Function "chfl_cell", at cell.h:33
    c_lib.chfl_cell.argtypes = [chfl_vector3d]
    c_lib.chfl_cell.restype = POINTER(CHFL_CELL)

    # Function "chfl_cell_triclinic", at cell.h:50
    c_lib.chfl_cell_triclinic.argtypes = [chfl_vector3d, chfl_vector3d]
    c_lib.chfl_cell_triclinic.restype = POINTER(CHFL_CELL)

    # Function "chfl_cell_from_frame", at cell.h:72
    c_lib.chfl_cell_from_frame.argtypes = [Frame]
    c_lib.chfl_cell_from_frame.restype = POINTER(CHFL_CELL)

    # Function "chfl_cell_copy", at cell.h:82
    c_lib.chfl_cell_copy.argtypes = [UnitCell]
    c_lib.chfl_cell_copy.restype = POINTER(CHFL_CELL)

    # Function "chfl_cell_volume", at cell.h:89
    c_lib.chfl_cell_volume.argtypes = [UnitCell, POINTER(c_double)]
    c_lib.chfl_cell_volume.restype = chfl_status
    c_lib.chfl_cell_volume.errcheck = _check_return_code

    # Function "chfl_cell_lengths", at cell.h:98
    c_lib.chfl_cell_lengths.argtypes = [UnitCell, chfl_vector3d]
    c_lib.chfl_cell_lengths.restype = chfl_status
    c_lib.chfl_cell_lengths.errcheck = _check_return_code

    # Function "chfl_cell_set_lengths", at cell.h:109
    c_lib.chfl_cell_set_lengths.argtypes = [UnitCell, chfl_vector3d]
    c_lib.chfl_cell_set_lengths.restype = chfl_status
    c_lib.chfl_cell_set_lengths.errcheck = _check_return_code

    # Function "chfl_cell_angles", at cell.h:118
    c_lib.chfl_cell_angles.argtypes = [UnitCell, chfl_vector3d]
    c_lib.chfl_cell_angles.restype = chfl_status
    c_lib.chfl_cell_angles.errcheck = _check_return_code

    # Function "chfl_cell_set_angles", at cell.h:131
    c_lib.chfl_cell_set_angles.argtypes = [UnitCell, chfl_vector3d]
    c_lib.chfl_cell_set_angles.restype = chfl_status
    c_lib.chfl_cell_set_angles.errcheck = _check_return_code

    # Function "chfl_cell_matrix", at cell.h:149
    c_lib.chfl_cell_matrix.argtypes = [UnitCell, ARRAY(chfl_vector3d, (3))]
    c_lib.chfl_cell_matrix.restype = chfl_status
    c_lib.chfl_cell_matrix.errcheck = _check_return_code

    # Function "chfl_cell_shape", at cell.h:158
    c_lib.chfl_cell_shape.argtypes = [UnitCell, POINTER(chfl_cellshape)]
    c_lib.chfl_cell_shape.restype = chfl_status
    c_lib.chfl_cell_shape.errcheck = _check_return_code

    # Function "chfl_cell_set_shape", at cell.h:167
    c_lib.chfl_cell_set_shape.argtypes = [UnitCell, chfl_cellshape]
    c_lib.chfl_cell_set_shape.restype = chfl_status
    c_lib.chfl_cell_set_shape.errcheck = _check_return_code

    # Function "chfl_cell_wrap", at cell.h:176
    c_lib.chfl_cell_wrap.argtypes = [UnitCell, chfl_vector3d]
    c_lib.chfl_cell_wrap.restype = chfl_status
    c_lib.chfl_cell_wrap.errcheck = _check_return_code

    # Function "chfl_frame", at frame.h:20
    c_lib.chfl_frame.argtypes = []
    c_lib.chfl_frame.restype = POINTER(CHFL_FRAME)

    # Function "chfl_frame_copy", at frame.h:30
    c_lib.chfl_frame_copy.argtypes = [Frame]
    c_lib.chfl_frame_copy.restype = POINTER(CHFL_FRAME)

    # Function "chfl_frame_atoms_count", at frame.h:38
    c_lib.chfl_frame_atoms_count.argtypes = [Frame, POINTER(c_uint64)]
    c_lib.chfl_frame_atoms_count.restype = chfl_status
    c_lib.chfl_frame_atoms_count.errcheck = _check_return_code

    # Function "chfl_frame_positions", at frame.h:57
    c_lib.chfl_frame_positions.argtypes = [Frame, POINTER(POINTER(chfl_vector3d)), POINTER(c_uint64)]
    c_lib.chfl_frame_positions.restype = chfl_status
    c_lib.chfl_frame_positions.errcheck = _check_return_code

    # Function "chfl_frame_velocities", at frame.h:80
    c_lib.chfl_frame_velocities.argtypes = [Frame, POINTER(POINTER(chfl_vector3d)), POINTER(c_uint64)]
    c_lib.chfl_frame_velocities.restype = chfl_status
    c_lib.chfl_frame_velocities.errcheck = _check_return_code

    # Function "chfl_frame_add_atom", at frame.h:92
    c_lib.chfl_frame_add_atom.argtypes = [Frame, Atom, chfl_vector3d, chfl_vector3d]
    c_lib.chfl_frame_add_atom.restype = chfl_status
    c_lib.chfl_frame_add_atom.errcheck = _check_return_code

    # Function "chfl_frame_remove", at frame.h:105
    c_lib.chfl_frame_remove.argtypes = [Frame, c_uint64]
    c_lib.chfl_frame_remove.restype = chfl_status
    c_lib.chfl_frame_remove.errcheck = _check_return_code

    # Function "chfl_frame_resize", at frame.h:117
    c_lib.chfl_frame_resize.argtypes = [Frame, c_uint64]
    c_lib.chfl_frame_resize.restype = chfl_status
    c_lib.chfl_frame_resize.errcheck = _check_return_code

    # Function "chfl_frame_add_velocities", at frame.h:129
    c_lib.chfl_frame_add_velocities.argtypes = [Frame]
    c_lib.chfl_frame_add_velocities.restype = chfl_status
    c_lib.chfl_frame_add_velocities.errcheck = _check_return_code

    # Function "chfl_frame_has_velocities", at frame.h:137
    c_lib.chfl_frame_has_velocities.argtypes = [Frame, POINTER(c_bool)]
    c_lib.chfl_frame_has_velocities.restype = chfl_status
    c_lib.chfl_frame_has_velocities.errcheck = _check_return_code

    # Function "chfl_frame_set_cell", at frame.h:146
    c_lib.chfl_frame_set_cell.argtypes = [Frame, UnitCell]
    c_lib.chfl_frame_set_cell.restype = chfl_status
    c_lib.chfl_frame_set_cell.errcheck = _check_return_code

    # Function "chfl_frame_set_topology", at frame.h:158
    c_lib.chfl_frame_set_topology.argtypes = [Frame, Topology]
    c_lib.chfl_frame_set_topology.restype = chfl_status
    c_lib.chfl_frame_set_topology.errcheck = _check_return_code

    # Function "chfl_frame_step", at frame.h:168
    c_lib.chfl_frame_step.argtypes = [Frame, POINTER(c_uint64)]
    c_lib.chfl_frame_step.restype = chfl_status
    c_lib.chfl_frame_step.errcheck = _check_return_code

    # Function "chfl_frame_set_step", at frame.h:177
    c_lib.chfl_frame_set_step.argtypes = [Frame, c_uint64]
    c_lib.chfl_frame_set_step.restype = chfl_status
    c_lib.chfl_frame_set_step.errcheck = _check_return_code

    # Function "chfl_frame_guess_bonds", at frame.h:189
    c_lib.chfl_frame_guess_bonds.argtypes = [Frame]
    c_lib.chfl_frame_guess_bonds.restype = chfl_status
    c_lib.chfl_frame_guess_bonds.errcheck = _check_return_code

    # Function "chfl_frame_distance", at frame.h:198
    c_lib.chfl_frame_distance.argtypes = [Frame, c_uint64, c_uint64, POINTER(c_double)]
    c_lib.chfl_frame_distance.restype = chfl_status
    c_lib.chfl_frame_distance.errcheck = _check_return_code

    # Function "chfl_frame_angle", at frame.h:209
    c_lib.chfl_frame_angle.argtypes = [Frame, c_uint64, c_uint64, c_uint64, POINTER(c_double)]
    c_lib.chfl_frame_angle.restype = chfl_status
    c_lib.chfl_frame_angle.errcheck = _check_return_code

    # Function "chfl_frame_dihedral", at frame.h:220
    c_lib.chfl_frame_dihedral.argtypes = [Frame, c_uint64, c_uint64, c_uint64, c_uint64, POINTER(c_double)]
    c_lib.chfl_frame_dihedral.restype = chfl_status
    c_lib.chfl_frame_dihedral.errcheck = _check_return_code

    # Function "chfl_frame_out_of_plane", at frame.h:234
    c_lib.chfl_frame_out_of_plane.argtypes = [Frame, c_uint64, c_uint64, c_uint64, c_uint64, POINTER(c_double)]
    c_lib.chfl_frame_out_of_plane.restype = chfl_status
    c_lib.chfl_frame_out_of_plane.errcheck = _check_return_code

    # Function "chfl_frame_properties_count", at frame.h:243
    c_lib.chfl_frame_properties_count.argtypes = [Frame, POINTER(c_uint64)]
    c_lib.chfl_frame_properties_count.restype = chfl_status
    c_lib.chfl_frame_properties_count.errcheck = _check_return_code

    # Function "chfl_frame_list_properties", at frame.h:259
    c_lib.chfl_frame_list_properties.argtypes = [Frame, POINTER(c_char_p), c_uint64]
    c_lib.chfl_frame_list_properties.restype = chfl_status
    c_lib.chfl_frame_list_properties.errcheck = _check_return_code

    # Function "chfl_frame_set_property", at frame.h:271
    c_lib.chfl_frame_set_property.argtypes = [Frame, c_char_p, Property]
    c_lib.chfl_frame_set_property.restype = chfl_status
    c_lib.chfl_frame_set_property.errcheck = _check_return_code

    # Function "chfl_frame_get_property", at frame.h:285
    c_lib.chfl_frame_get_property.argtypes = [Frame, c_char_p]
    c_lib.chfl_frame_get_property.restype = POINTER(CHFL_PROPERTY)

    # Function "chfl_frame_add_bond", at frame.h:294
    c_lib.chfl_frame_add_bond.argtypes = [Frame, c_uint64, c_uint64]
    c_lib.chfl_frame_add_bond.restype = chfl_status
    c_lib.chfl_frame_add_bond.errcheck = _check_return_code

    # Function "chfl_frame_bond_with_order", at frame.h:304
    c_lib.chfl_frame_bond_with_order.argtypes = [Frame, c_uint64, c_uint64, chfl_bond_order]
    c_lib.chfl_frame_bond_with_order.restype = chfl_status
    c_lib.chfl_frame_bond_with_order.errcheck = _check_return_code

    # Function "chfl_frame_remove_bond", at frame.h:316
    c_lib.chfl_frame_remove_bond.argtypes = [Frame, c_uint64, c_uint64]
    c_lib.chfl_frame_remove_bond.restype = chfl_status
    c_lib.chfl_frame_remove_bond.errcheck = _check_return_code

    # Function "chfl_frame_add_residue", at frame.h:328
    c_lib.chfl_frame_add_residue.argtypes = [Frame, Residue]
    c_lib.chfl_frame_add_residue.restype = chfl_status
    c_lib.chfl_frame_add_residue.errcheck = _check_return_code

    # Function "chfl_trajectory_open", at trajectory.h:22
    c_lib.chfl_trajectory_open.argtypes = [c_char_p, c_char]
    c_lib.chfl_trajectory_open.restype = POINTER(CHFL_TRAJECTORY)

    # Function "chfl_trajectory_with_format", at trajectory.h:39
    c_lib.chfl_trajectory_with_format.argtypes = [c_char_p, c_char, c_char_p]
    c_lib.chfl_trajectory_with_format.restype = POINTER(CHFL_TRAJECTORY)

    # Function "chfl_trajectory_path", at trajectory.h:52
    c_lib.chfl_trajectory_path.argtypes = [Trajectory, POINTER(c_char_p)]
    c_lib.chfl_trajectory_path.restype = chfl_status
    c_lib.chfl_trajectory_path.errcheck = _check_return_code

    # Function "chfl_trajectory_read", at trajectory.h:64
    c_lib.chfl_trajectory_read.argtypes = [Trajectory, Frame]
    c_lib.chfl_trajectory_read.restype = chfl_status
    c_lib.chfl_trajectory_read.errcheck = _check_return_code

    # Function "chfl_trajectory_read_step", at trajectory.h:76
    c_lib.chfl_trajectory_read_step.argtypes = [Trajectory, c_uint64, Frame]
    c_lib.chfl_trajectory_read_step.restype = chfl_status
    c_lib.chfl_trajectory_read_step.errcheck = _check_return_code

    # Function "chfl_trajectory_write", at trajectory.h:85
    c_lib.chfl_trajectory_write.argtypes = [Trajectory, Frame]
    c_lib.chfl_trajectory_write.restype = chfl_status
    c_lib.chfl_trajectory_write.errcheck = _check_return_code

    # Function "chfl_trajectory_set_topology", at trajectory.h:96
    c_lib.chfl_trajectory_set_topology.argtypes = [Trajectory, Topology]
    c_lib.chfl_trajectory_set_topology.restype = chfl_status
    c_lib.chfl_trajectory_set_topology.errcheck = _check_return_code

    # Function "chfl_trajectory_topology_file", at trajectory.h:110
    c_lib.chfl_trajectory_topology_file.argtypes = [Trajectory, c_char_p, c_char_p]
    c_lib.chfl_trajectory_topology_file.restype = chfl_status
    c_lib.chfl_trajectory_topology_file.errcheck = _check_return_code

    # Function "chfl_trajectory_set_cell", at trajectory.h:120
    c_lib.chfl_trajectory_set_cell.argtypes = [Trajectory, UnitCell]
    c_lib.chfl_trajectory_set_cell.restype = chfl_status
    c_lib.chfl_trajectory_set_cell.errcheck = _check_return_code

    # Function "chfl_trajectory_nsteps", at trajectory.h:130
    c_lib.chfl_trajectory_nsteps.argtypes = [Trajectory, POINTER(c_uint64)]
    c_lib.chfl_trajectory_nsteps.restype = chfl_status
    c_lib.chfl_trajectory_nsteps.errcheck = _check_return_code

    # Function "chfl_selection", at selection.h:20
    c_lib.chfl_selection.argtypes = [c_char_p]
    c_lib.chfl_selection.restype = POINTER(CHFL_SELECTION)

    # Function "chfl_selection_copy", at selection.h:33
    c_lib.chfl_selection_copy.argtypes = [Selection]
    c_lib.chfl_selection_copy.restype = POINTER(CHFL_SELECTION)

    # Function "chfl_selection_size", at selection.h:45
    c_lib.chfl_selection_size.argtypes = [Selection, POINTER(c_uint64)]
    c_lib.chfl_selection_size.restype = chfl_status
    c_lib.chfl_selection_size.errcheck = _check_return_code

    # Function "chfl_selection_string", at selection.h:58
    c_lib.chfl_selection_string.argtypes = [Selection, c_char_p, c_uint64]
    c_lib.chfl_selection_string.restype = chfl_status
    c_lib.chfl_selection_string.errcheck = _check_return_code

    # Function "chfl_selection_evaluate", at selection.h:71
    c_lib.chfl_selection_evaluate.argtypes = [Selection, Frame, POINTER(c_uint64)]
    c_lib.chfl_selection_evaluate.restype = chfl_status
    c_lib.chfl_selection_evaluate.errcheck = _check_return_code

    # Function "chfl_selection_matches", at selection.h:97
    c_lib.chfl_selection_matches.argtypes = [Selection, ndpointer(chfl_match, flags="C_CONTIGUOUS", ndim=1), c_uint64]
    c_lib.chfl_selection_matches.restype = chfl_status
    c_lib.chfl_selection_matches.errcheck = _check_return_code



#!/usr/bin/env python
# -* coding: utf-8 -*

"""
This python script use Blue Obelisk's elements.xml file to create the
corresponding C++ header with atomic information.

The elements.xml file can be found at the official SVN repo:
    http://svn.code.sf.net/p/bodr/code/trunk/bodr
"""

import sys
from xml.etree import ElementTree as ET


class Atom:
    def __init__(self, symbol, name, number, mass, cov, VdW):
        self.symbol = symbol
        self.name = name
        self.number = number
        self.mass = mass
        self.cov = cov
        self.VdW = VdW

    def __str__(self):
        return '{{"{}", {{{}, std::string("{}"), {}, nullopt, {}, {}}}}}'.format(
            self.symbol, self.number, self.name, self.mass, self.cov, self.VdW
        )


def read_elements(path):
    root = ET.parse(path).getroot()
    atoms = []

    for atom in root[1:]:
        for prop in atom:
            if prop.get("dictRef") == "bo:name":
                name = prop.get("value")
            if prop.get("dictRef") == "bo:symbol":
                symbol = prop.get("value")
            if prop.get("dictRef") == "bo:atomicNumber":
                number = int(prop.text)
            if prop.get("dictRef") == "bo:mass":
                assert prop.get("units") == "units:atmass"
                mass = float(prop.text)
            if prop.get("dictRef") == "bo:radiusCovalent":
                assert prop.get("units") == "units:ang"
                cov = float(prop.text)
            if prop.get("dictRef") == "bo:radiusVDW":
                assert prop.get("units") == "units:ang"
                VdW = float(prop.text)
        atoms.append(Atom(symbol, name, number, mass, cov, VdW))

    return atoms


HEADER = """// Chemfiles, a modern library for chemistry file reading and writing
// Copyright (C) Guillaume Fraux and contributors -- BSD license

// !!!! AUTO-GENERATED FILE !!!! Do not edit. See elements.py for the code.
// The data comes from Blue Obelisk's data repository at the svn repository:
// http://svn.code.sf.net/p/bodr/code/trunk/bodr

#include "chemfiles/periodic_table.hpp"
using namespace chemfiles;
"""

ARRAY = """
const atomic_data_map chemfiles::PERIODIC_TABLE = {
"""


def write_elements(path, elements):
    with open(path, "w") as fd:
        fd.write(HEADER)
        fd.write(ARRAY)
        for atom in atoms:
            fd.write("    " + str(atom) + ",\n")
        fd.write("};\n")


def usage():
    print(sys.argv[0] + " path/to/elements.xml periodic_table.cpp")


if __name__ == "__main__":
    if len(sys.argv) < 3:
        usage()
        sys.exit(0)

    atoms = read_elements(sys.argv[1])
    write_elements(sys.argv[2], atoms)



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from .misc import ChemfilesError, set_warnings_callback, add_configuration
from .atom import Atom
from .residue import Residue
from .topology import Topology, BondOrder
from .cell import UnitCell, CellShape
from .frame import Frame
from .trajectory import Trajectory
from .selection import Selection
from .property import Property

__version__ = "0.9.1"



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import sys
import numpy as np
from ctypes import c_double, c_bool

from .ffi import chfl_property_kind, chfl_vector3d
from .misc import ChemfilesError
from ._utils import CxxPointer, _call_with_growing_buffer, string_type


class Property(CxxPointer):
    """
    A :py:class:`Property` holds the data used in properties in
    :py:class:`Frame` and :py:class:`Atom`. A property can have various types:
    bool, double, string or 3D vectors.

    This class is not meant for direct use, but is an internal class.
    """

    def __init__(self, value):
        """Create a new property containing the given value"""
        if isinstance(value, bool):
            ptr = self.ffi.chfl_property_bool(c_bool(value))
        elif isinstance(value, (float, int)):
            ptr = self.ffi.chfl_property_double(c_double(value))
        elif isinstance(value, string_type):
            ptr = self.ffi.chfl_property_string(value.encode("utf8"))
        elif _is_vector3d(value):
            value = chfl_vector3d(value[0], value[1], value[2])
            ptr = self.ffi.chfl_property_vector3d(value)
        else:
            raise ChemfilesError(
                "can not create a Property with a value of type '{}'".format(type(value))
            )

        super(Property, self).__init__(ptr, is_const=False)

    def get(self):
        kind = chfl_property_kind()
        self.ffi.chfl_property_get_kind(self, kind)
        if kind.value == chfl_property_kind.CHFL_PROPERTY_BOOL:
            value = c_bool()
            self.ffi.chfl_property_get_bool(self, value)
            return value.value
        if kind.value == chfl_property_kind.CHFL_PROPERTY_DOUBLE:
            value = c_double()
            self.ffi.chfl_property_get_double(self, value)
            return value.value
        if kind.value == chfl_property_kind.CHFL_PROPERTY_STRING:

            def callback(buffer, size):
                self.ffi.chfl_property_get_string(self, buffer, size)

            return _call_with_growing_buffer(callback, initial=32)
        if kind.value == chfl_property_kind.CHFL_PROPERTY_VECTOR3D:
            value = chfl_vector3d()
            self.ffi.chfl_property_get_vector3d(self, value)
            return value[0], value[1], value[2]
        else:
            raise ChemfilesError("unknown property kind, this is a bug")


def _is_vector3d(value):
    try:
        a = np.array(value, dtype="double")
        return len(a) >= 3
    except:
        return False



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import warnings

from .clib import _get_c_library


class ChemfilesWarning(UserWarning):
    """Warnings from the Chemfiles runtime."""

    pass


class ChemfilesError(BaseException):
    """Exception class for errors in chemfiles"""

    pass


# Store a reference to the last logging callback, to preven Python from
# garbage-collecting it.
_CURRENT_CALLBACK = None


def set_warnings_callback(function):
    """
    Call `function` on every warning event. The callback should take a string
    message and return nothing.

    By default, warnings are send to python `warnings` module.
    """
    from .ffi import chfl_warning_callback

    def callback(message):
        try:
            function(message.decode("utf8"))
        except Exception as e:
            message = "exception raised in warning callback: {}".format(e)
            warnings.warn(message, ChemfilesWarning)

    global _CURRENT_CALLBACK
    _CURRENT_CALLBACK = chfl_warning_callback(callback)

    _get_c_library().chfl_set_warning_callback(_CURRENT_CALLBACK)


def add_configuration(path):
    """
    Read configuration data from the file at ``path``.

    By default, chemfiles reads configuration from any file name `.chemfilesrc`
    in the current directory or any parent directory. This function can be used
    to add data from another configuration file.

    This function will fail if there is no file at ``path``, or if the file is
    incorectly formatted. Data from the new configuration file will overwrite
    any existing data.
    """
    _get_c_library().chfl_add_configuration(path.encode("utf8"))


def _last_error():
    """Get the last error from the chemfiles runtime."""
    return _get_c_library().chfl_last_error().decode("utf8")


def _clear_errors():
    """Clear any error message saved in the chemfiles runtime."""
    return _get_c_library().chfl_clear_errors()


def _set_default_warning_callback():
    set_warnings_callback(
        # We need to set stacklevel=4 to get through the lambda =>
        # adapatator => C++ code => Python binding => user code
        lambda message: warnings.warn(message, ChemfilesWarning, stacklevel=4)
    )



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import numpy as np
from ctypes import c_uint64, c_bool, c_double, c_char_p, POINTER

from ._utils import CxxPointer, string_type
from .misc import ChemfilesError
from .ffi import chfl_vector3d, chfl_bond_order
from .atom import Atom
from .topology import Topology
from .cell import UnitCell
from .property import Property


class FrameAtoms(object):
    """Proxy object to get the atoms in a frame"""

    def __init__(self, frame):
        self.frame = frame

    def __len__(self):
        """Get the current number of atoms in this :py:class:`Frame`."""
        count = c_uint64()
        self.frame.ffi.chfl_frame_atoms_count(self.frame, count)
        return count.value

    def __getitem__(self, index):
        """
        Get a reference to the :py:class:`Atom` at the given ``index`` in the
        associated :py:class:`Frame`.
        """
        if index >= len(self):
            raise IndexError("atom index ({}) out of range for this frame".format(index))
        else:
            ptr = self.frame.ffi.chfl_atom_from_frame(self.frame, c_uint64(index))
            return Atom.from_ptr(ptr)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __repr__(self):
        return "[" + ", ".join([atom.__repr__() for atom in self]) + "]"


class Frame(CxxPointer):
    """
    A :py:class:`Frame` contains data from one simulation step: the current
    unit cell, the topology, the positions, and the velocities of the particles
    in the system. If some information is missing (topology or velocity or unit
    cell), the corresponding data is filled with a default value.
    """

    def __init__(self):
        """
        Create an empty :py:class:`Frame` that will be resized by the runtime
        as needed.
        """
        super(Frame, self).__init__(self.ffi.chfl_frame(), is_const=False)

    def __copy__(self):
        return Frame.from_ptr(self.ffi.chfl_frame_copy(self))

    def __repr__(self):
        return "Frame with {} atoms".format(len(self.atoms))

    @property
    def atoms(self):
        return FrameAtoms(self)

    def resize(self, count):
        """
        Resize the positions, velocities and topology in this
        :py:class:`Frame`, to have space for `count` atoms.

        This function may invalidate any array of the positions or the
        velocities if the new size is bigger than the old one. In all cases,
        previous data is conserved. This function conserve the presence or
        absence of velocities.
        """
        self.ffi.chfl_frame_resize(self, c_uint64(count))

    def add_atom(self, atom, position, velocity=None):
        """
        Add a copy of the :py:class:`Atom` ``atom`` and the corresponding
        ``position`` and ``velocity`` to this :py:class:`Frame`.

        ``velocity`` can be ``None`` if no velocity is associated with the
        atom.
        """
        position = chfl_vector3d(position[0], position[1], position[2])
        if velocity:
            velocity = chfl_vector3d(velocity[0], velocity[1], velocity[2])
        self.ffi.chfl_frame_add_atom(self, atom, position, velocity)

    def remove(self, i):
        """
        Remove the atom at index ``i`` in this :py:class:`Frame`.

        This modify all the atoms indexes after ``i``, and invalidate any array
        obtained using :py:func:`Frame.positions` or
        :py:func:`Frame.velocities`.
        """
        self.ffi.chfl_frame_remove(self, c_uint64(i))

    def add_bond(self, i, j, order=None):
        """
        Add a bond between the atoms at indexes ``i`` and ``j`` in this
        :py:class:`Frame`'s topology, optionally setting the bond ``order``.
        """
        if order is None:
            self.ffi.chfl_frame_add_bond(self, c_uint64(i), c_uint64(j))
        else:
            self.ffi.chfl_frame_bond_with_order(
                self, c_uint64(i), c_uint64(j), chfl_bond_order(order)
            )

    def remove_bond(self, i, j):
        """
        Remove any existing bond between the atoms at indexes ``i`` and ``j``
        in this :py:class:`Frame`'s topology.

        This function does nothing if there is no bond between ``i`` and ``j``.
        """
        self.ffi.chfl_frame_remove_bond(self, c_uint64(i), c_uint64(j))

    def add_residue(self, residue):
        """
        Add the :py:class:`Residue` ``residue`` to this :py:class:`Frame`'s
        topology.

        The residue ``id`` must not already be in the topology, and the residue
        must contain only atoms that are not already in another residue.
        """
        self.ffi.chfl_frame_add_residue(self, residue)

    @property
    def positions(self):
        """
        Get a view into the positions of this :py:class:`Frame`.

        This function gives direct access to the positions as a numpy array.
        Modifying the array will change the positions in the frame.

        If the frame is resized (by writing to it, calling
        :py:func:`Frame.resize`, :py:func:`Frame.add_atom`,
        :py:func:`Frame.remove`), the array is invalidated. Accessing it can
        cause a segfault.
        """
        count = c_uint64()
        data = POINTER(chfl_vector3d)()
        self.ffi.chfl_frame_positions(self, data, count)
        count = count.value
        if count != 0:
            positions = np.ctypeslib.as_array(data, shape=(count,))
            return positions.view(np.float64).reshape((count, 3))
        else:
            return np.array([[], [], []], dtype=np.float64)

    @property
    def velocities(self):
        """
        Get a view into the velocities of this :py:class:`Frame`.

        This function gives direct access to the velocities as a numpy array.
        Modifying the array will change the velocities in the frame.

        If the frame is resized (by writing to it, calling
        :py:func:`Frame.resize`, :py:func:`Frame.add_atom`,
        :py:func:`Frame.remove`), the array is invalidated. Accessing it can
        cause a segfault.
        """
        count = c_uint64()
        data = POINTER(chfl_vector3d)()
        self.ffi.chfl_frame_velocities(self, data, count)
        count = count.value
        if count != 0:
            velocities = np.ctypeslib.as_array(data, shape=(count,))
            return velocities.view(np.float64).reshape((count, 3))
        else:
            return np.array([[], [], []], dtype=np.float64)

    def add_velocities(self):
        """
        Add velocity data to this :py:class:`Frame`.

        The velocities are initialized to zero. If the frame already contains
        velocities, this function does nothing.
        """
        self.ffi.chfl_frame_add_velocities(self)

    def has_velocities(self):
        """Check if this :py:class:`Frame` contains velocity."""
        velocities = c_bool()
        self.ffi.chfl_frame_has_velocities(self, velocities)
        return velocities.value

    @property
    def cell(self):
        """
        Get a mutable reference to the :py:class:`UnitCell` of this
        :py:class:`Frame`. Any modification to the cell will be reflected in
        the frame.
        """
        return UnitCell.from_ptr(self.ffi.chfl_cell_from_frame(self))

    @cell.setter
    def cell(self, cell):
        """
        Set the :py:class:`UnitCell` of this :py:class:`Frame` to ``cell``.
        """
        self.ffi.chfl_frame_set_cell(self, cell)

    @property
    def topology(self):
        """
        Get read-only access to the :py:class:`Topology` of this
        :py:class:`Frame`.
        """
        return Topology.from_const_ptr(self.ffi.chfl_topology_from_frame(self))

    @topology.setter
    def topology(self, topology):
        """
        Set the :py:class:`Topology` of this :py:class:`Frame` to ``topology``.
        """
        self.ffi.chfl_frame_set_topology(self, topology)

    @property
    def step(self):
        """
        Get this :py:class:`Frame` step, i.e. the frame number in the
        trajectory.
        """
        step = c_uint64()
        self.ffi.chfl_frame_step(self, step)
        return step.value

    @step.setter
    def step(self, step):
        """Set this :py:class:`Frame` step to ``step``."""
        self.ffi.chfl_frame_set_step(self, c_uint64(step))

    def guess_bonds(self):
        """
        Guess the bonds, angles and dihedrals in this :py:class:`Frame`.

        The bonds are guessed using a distance-based algorithm, and then angles
        and dihedrals are guessed from the bonds.
        """
        self.ffi.chfl_frame_guess_bonds(self)

    def distance(self, i, j):
        """
        Get the distance between the atoms at indexes ``i`` and ``j`` in this
        :py:class:`Frame`, accounting for periodic boundary conditions. The
        result is expressed in angstroms.
        """
        distance = c_double()
        self.ffi.chfl_frame_distance(self, c_uint64(i), c_uint64(j), distance)
        return distance.value

    def angle(self, i, j, k):
        """
        Get the angle formed by the atoms at indexes ``i``, ``j`` and ``k`` in
        this :py:class:`Frame`, accounting for periodic boundary conditions.
        The result is expressed in radians.
        """
        angle = c_double()
        self.ffi.chfl_frame_angle(self, c_uint64(i), c_uint64(j), c_uint64(k), angle)
        return angle.value

    def dihedral(self, i, j, k, m):
        """
        Get the dihedral angle formed by the atoms at indexes ``i``, ``j``,
        ``k`` and ``m`` in this :py:class:`Frame`, accounting for periodic
        boundary conditions. The result is expressed in radians.
        """
        dihedral = c_double()
        self.ffi.chfl_frame_dihedral(
            self, c_uint64(i), c_uint64(j), c_uint64(k), c_uint64(m), dihedral
        )
        return dihedral.value

    def out_of_plane(self, i, j, k, m):
        """
        Get the out of plane distance formed by the atoms at indexes ``i``,
        ``j``, ``k`` and ``m`` in this :py:class:`Frame`, accounting for
        periodic boundary conditions. The result is expressed in angstroms.

        This is the distance betweent the atom j and the ikm plane. The j atom
        is the center of the improper dihedral angle formed by i, j, k and m.
        """
        distance = c_double()
        self.ffi.chfl_frame_out_of_plane(
            self, c_uint64(i), c_uint64(j), c_uint64(k), c_uint64(m), distance
        )
        return distance.value

    def __iter__(self):
        # Disable automatic iteration from __getitem__
        raise TypeError("use Frame.atoms to iterate over a frame")

    def __getitem__(self, name):
        """
        Get a property of this frame with the given ``name``, or raise an error
        if the property does not exists.
        """
        if not isinstance(name, string_type):
            raise ChemfilesError(
                "Invalid type {} for a frame property name".format(type(name))
            )
        ptr = self.ffi.chfl_frame_get_property(self, name.encode("utf8"))
        return Property.from_ptr(ptr).get()

    def __setitem__(self, name, value):
        """
        Set a property of this frame, with the given ``name`` and ``value``.
        The new value overwrite any pre-existing property with the same name.
        """
        if not isinstance(name, string_type):
            raise ChemfilesError(
                "Invalid type {} for a frame property name".format(type(name))
            )
        self.ffi.chfl_frame_set_property(self, name.encode("utf8"), Property(value))

    def properties_count(self):
        """Get the number of properties in this frame."""
        count = c_uint64()
        self.ffi.chfl_frame_properties_count(self, count)
        return count.value

    def list_properties(self):
        """Get the name of all properties in this frame."""
        count = self.properties_count()
        StringArray = c_char_p * count
        names = StringArray()
        self.ffi.chfl_frame_list_properties(self, names, count)
        return list(map(lambda n: n.decode("utf8"), names))



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from ctypes import c_double, ARRAY
from enum import IntEnum

from ._utils import CxxPointer
from .ffi import chfl_cellshape, chfl_vector3d


class CellShape(IntEnum):
    """
    Available cell shapes in Chemfiles:

    - ``CellType.Orthorhombic``: for cells where the three angles are 90°;
    - ``CellType.Triclinic``: for cells where the three angles may not be 90°;
    - ``CellType.Infinite``: for cells without periodic boundary conditions;
    """

    Orthorhombic = chfl_cellshape.CHFL_CELL_ORTHORHOMBIC
    Triclinic = chfl_cellshape.CHFL_CELL_TRICLINIC
    Infinite = chfl_cellshape.CHFL_CELL_INFINITE


class UnitCell(CxxPointer):
    """
    An :py:class:`UnitCell` represent the box containing the atoms, and its
    periodicity.

    An unit cell is fully represented by three lenghts (a, b, c); and three
    angles (alpha, beta, gamma). The angles are stored in degrees, and the
    lenghts in Angstroms. The cell angles are defined as follow: alpha is the
    angles between the cell vectors `b` and `c`; beta as the angle between `a`
    and `c`; and gamma as the angle between `a` and `b`.

    A cell also has a matricial representation, by projecting the three base
    vector into an orthonormal base. We choose to represent such matrix as an
    upper triangular matrix:

    .. code-block:: sh

        | a_x   b_x   c_x |
        |  0    b_y   c_y |
        |  0     0    c_z |
    """

    def __init__(self, a, b, c, alpha=90.0, beta=90.0, gamma=90.0):
        """
        Create a new :py:class:`UnitCell` with cell lenghts of ``a``, ``b`` and
        ``c``, and cell angles ``alpha``, ``beta`` and ``gamma``.

        If alpha, beta and gamma are equal to 90.0, the new unit cell shape is
        ``CellShape.Orthorhombic``. Else it is ``CellShape.Infinite``.
        """
        lenghts = chfl_vector3d(a, b, c)
        angles = chfl_vector3d(alpha, beta, gamma)
        if alpha == 90.0 and beta == 90.0 and gamma == 90.0:
            ptr = self.ffi.chfl_cell(lenghts)
        else:
            ptr = self.ffi.chfl_cell_triclinic(lenghts, angles)
        super(UnitCell, self).__init__(ptr, is_const=False)

    def __copy__(self):
        return UnitCell.from_ptr(self.ffi.chfl_cell_copy(self))

    def __repr__(self):
        return "UnitCell({}, {}, {}, {}, {}, {})".format(*(self.lengths + self.angles))

    @property
    def lengths(self):
        """Get the three lenghts of this :py:class:`UnitCell`, in Angstroms."""
        lengths = chfl_vector3d(0, 0, 0)
        self.ffi.chfl_cell_lengths(self, lengths)
        return lengths[0], lengths[1], lengths[2]

    @lengths.setter
    def lengths(self, lengths):
        """
        Set the three lenghts of this :py:class:`UnitCell` to ``lengths``. The
        values should be in Angstroms.
        """
        a, b, c = lengths
        self.ffi.chfl_cell_set_lengths(self, chfl_vector3d(a, b, c))

    @property
    def angles(self):
        """Get the three angles of this :py:class:`UnitCell`, in degrees."""
        angles = chfl_vector3d(0, 0, 0)
        self.ffi.chfl_cell_angles(self, angles)
        return angles[0], angles[1], angles[2]

    @angles.setter
    def angles(self, angles):
        """
        Set the three angles of this :py:class:`UnitCell` to ``alpha``,
        ``beta`` and ``gamma``. These values should be in degrees. Setting
        angles is only possible for ``CellShape.Triclinic`` cells.
        """
        alpha, beta, gamma = angles
        self.ffi.chfl_cell_set_angles(self, chfl_vector3d(alpha, beta, gamma))

    @property
    def matrix(self):
        """
        Get this :py:class:`UnitCell` matricial representation.

        The matricial representation is obtained by aligning the a vector along
        the *x* axis and putting the b vector in the *xy* plane. This make the
        matrix an upper triangular matrix:

        .. code-block:: sh

            | a_x   b_x   c_x |
            |  0    b_y   c_y |
            |  0     0    c_z |
        """
        m = ARRAY(chfl_vector3d, 3)()
        self.ffi.chfl_cell_matrix(self, m)
        return [
            (m[0][0], m[0][1], m[0][2]),
            (m[1][0], m[1][1], m[1][2]),
            (m[2][0], m[2][1], m[2][2]),
        ]

    @property
    def shape(self):
        """Get the shape of this :py:class:`UnitCell`."""
        shape = chfl_cellshape()
        self.ffi.chfl_cell_shape(self, shape)
        return CellShape(shape.value)

    @shape.setter
    def shape(self, shape):
        """Set the shape of this :py:class:`UnitCell` to ``shape``."""
        self.ffi.chfl_cell_set_shape(self, chfl_cellshape(shape))

    @property
    def volume(self):
        """Get the volume of this :py:class:`UnitCell`."""
        volume = c_double()
        self.ffi.chfl_cell_volume(self, volume)
        return volume.value

    def wrap(self, vector):
        """
        Wrap a ``vector`` in this :py:class:`UnitCell`, and return the wrapped
        vector.
        """
        vector = chfl_vector3d(vector[0], vector[1], vector[2])
        self.ffi.chfl_cell_wrap(self, vector)
        return (vector[0], vector[1], vector[2])



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from ctypes import c_uint64, c_bool
from enum import IntEnum
import numpy as np

from ._utils import CxxPointer
from .ffi import chfl_bond_order
from .atom import Atom
from .residue import Residue


class BondOrder(IntEnum):
    """
    Possible bond orders are:

    - ``BondOrder.Unknown``: the bond order is not specified
    - ``BondOrder.Single``: bond order for single bond
    - ``BondOrder.Double``: bond order for double bond
    - ``BondOrder.Triple``: bond order for triple bond
    - ``BondOrder.Quadruple``: bond order for quadruple bond (present in some metals)
    - ``BondOrder.Qintuplet``: bond order for qintuplet bond (present in some metals)
    - ``BondOrder.Amide``: bond order for amide bond
    - ``BondOrder.Aromatic``: bond order for aromatic bond
    """

    Unknown = chfl_bond_order.CHFL_BOND_UNKNOWN
    Single = chfl_bond_order.CHFL_BOND_SINGLE
    Double = chfl_bond_order.CHFL_BOND_DOUBLE
    Triple = chfl_bond_order.CHFL_BOND_TRIPLE
    Quadruple = chfl_bond_order.CHFL_BOND_QUADRUPLE
    Qintuplet = chfl_bond_order.CHFL_BOND_QINTUPLET
    Amide = chfl_bond_order.CHFL_BOND_AMIDE
    Aromatic = chfl_bond_order.CHFL_BOND_AROMATIC


class TopologyAtoms(object):
    """Proxy object to get the atoms in a topology"""

    def __init__(self, topology):
        self.topology = topology

    def __len__(self):
        """Get the current number of atoms in this :py:class:`Topology`."""
        count = c_uint64()
        self.topology.ffi.chfl_topology_atoms_count(self.topology, count)
        return count.value

    def __getitem__(self, index):
        """
        Get a reference to the :py:class:`Atom` at the given ``index`` in the
        associated :py:class:`Topology`.
        """
        if index >= len(self):
            raise IndexError("atom index ({}) out of range for this topology".format(index))
        else:
            ptr = self.topology.ffi.chfl_atom_from_topology(self.topology, c_uint64(index))
            return Atom.from_ptr(ptr)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __delitem__(self, index):
        self.remove(index)

    def __repr__(self):
        return "[" + ", ".join([atom.__repr__() for atom in self]) + "]"

    def remove(self, index):
        """
        Remove the :py:class:`Atom` atthe given ``index`` from the associated
        :py:class:`Topology`.

        This shifts all the atoms indexes after ``i`` by 1 (n becomes n-1).
        """
        self.topology.ffi.chfl_topology_remove(self.topology, c_uint64(index))

    def append(self, atom):
        """
        Add a copy of the :py:class:`Atom` ``atom`` at the end of this
        :py:class:`Topology`.
        """
        self.topology.ffi.chfl_topology_add_atom(self.topology, atom)


class TopologyResidues(object):
    """Proxy object to get the residues in a topology"""

    def __init__(self, topology):
        self.topology = topology

    def __len__(self):
        """Get the current number of residues in this :py:class:`Topology`."""
        count = c_uint64()
        self.topology.ffi.chfl_topology_residues_count(self.topology, count)
        return count.value

    def __getitem__(self, index):
        """
        Get read-only access to the :py:class:`Residue` at the given ``index``
        from the associated :py:class:`Topology`. The residue index in the
        topology does not necessarily match the residue id.
        """
        if index >= len(self):
            raise IndexError("residue index ({}) out of range for this topology".format(index))
        else:
            ptr = self.topology.ffi.chfl_residue_from_topology(self.topology, c_uint64(index))
            return Residue.from_const_ptr(ptr)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __repr__(self):
        return "[" + ", ".join([residue.__repr__() for residue in self]) + "]"

    def append(self, residue):
        """
        Add the :py:class:`Residue` ``residue`` to this :py:class:`Topology`.

        The residue ``id`` must not already be in the topology, and the residue
        must contain only atoms that are not already in another residue.
        """
        self.topology.ffi.chfl_topology_add_residue(self.topology, residue)


class Topology(CxxPointer):
    """
    A :py:class:`Topology` contains the definition of all the atoms in the
    system, and the liaisons between the atoms (bonds, angles, dihedrals, ...).
    It will also contain all the residues information if it is available.
    """

    def __init__(self):
        """Create a new empty :py:class:`Topology`."""
        super(Topology, self).__init__(self.ffi.chfl_topology(), is_const=False)

    def __copy__(self):
        return Topology.from_ptr(self.ffi.chfl_topology_copy(self))

    def __repr__(self):
        return "Topology with {} atoms".format(len(self.atoms))

    @property
    def atoms(self):
        return TopologyAtoms(self)

    def resize(self, count):
        """
        Resize this :py:class:`Topology` to contain ``count`` atoms. If the
        new number of atoms is bigger than the current number, new atoms will
        be created with an empty name and type. If it is lower than the current
        number of atoms, the last atoms will be removed, together with the
        associated bonds, angles and dihedrals.
        """
        self.ffi.chfl_topology_resize(self, count)

    @property
    def residues(self):
        return TopologyResidues(self)

    def residue_for_atom(self, index):
        """
        Get read-only access to the :py:class:`Residue` containing the atom at
        the given ``index`` from this :py:class:`Topology`; or ``None`` if the
        atom is not part of a residue.
        """
        if index >= len(self.atoms):
            raise IndexError(
                "residue index ({}) out of range for this topology".format(index)
            )
        ptr = self.ffi.chfl_residue_for_atom(self, c_uint64(index))
        if ptr:
            return Residue.from_const_ptr(ptr)
        else:
            return None

    def residues_linked(self, first, second):
        """
        Check if the two :py:class:`Residue` ``first`` and ``second`` from this
        :py:class:`Topology` are linked together, *i.e.* if there is a bond
        between one atom in the first residue and one atom in the second one.
        """
        linked = c_bool()
        self.ffi.chfl_topology_residues_linked(self, first, second, linked)
        return linked.value

    def bonds_count(self):
        """Get the number of bonds in this :py:class:`Topology`."""
        bonds = c_uint64()
        self.ffi.chfl_topology_bonds_count(self, bonds)
        return bonds.value

    def angles_count(self):
        """Get the number of angles in this :py:class:`Topology`."""
        angles = c_uint64()
        self.ffi.chfl_topology_angles_count(self, angles)
        return angles.value

    def dihedrals_count(self):
        """Get the number of dihedral angles in this :py:class:`Topology`."""
        dihedrals = c_uint64()
        self.ffi.chfl_topology_dihedrals_count(self, dihedrals)
        return dihedrals.value

    def impropers_count(self):
        """Get the number of improper angles in this :py:class:`Topology`."""
        impropers = c_uint64()
        self.ffi.chfl_topology_impropers_count(self, impropers)
        return impropers.value

    @property
    def bonds(self):
        """Get the list of bonds in this :py:class:`Topology`."""
        count = self.bonds_count()
        bonds = np.zeros((count, 2), np.uint64)
        self.ffi.chfl_topology_bonds(self, bonds, c_uint64(count))
        return bonds

    def bonds_order(self, i, j):
        """
        Get the bonds order corresponding to the bond between atoms i and j
        """
        order = chfl_bond_order()
        self.ffi.chfl_topology_bond_order(self, c_uint64(i), c_uint64(j), order)
        return BondOrder(order.value)

    @property
    def bonds_orders(self):
        """
        Get the list of bonds order for each bond in this :py:class:`Topology`.
        """
        count = self.bonds_count()
        orders = np.zeros(count, chfl_bond_order)
        self.ffi.chfl_topology_bond_orders(self, orders, c_uint64(count))
        return list(map(BondOrder, orders))

    @property
    def angles(self):
        """Get the list of angles in this :py:class:`Topology`."""
        count = self.angles_count()
        angles = np.zeros((count, 3), np.uint64)
        self.ffi.chfl_topology_angles(self, angles, c_uint64(count))
        return angles

    @property
    def dihedrals(self):
        """Get the list of dihedral angles in this :py:class:`Topology`."""
        count = self.dihedrals_count()
        dihedrals = np.zeros((count, 4), np.uint64)
        self.ffi.chfl_topology_dihedrals(self, dihedrals, c_uint64(count))
        return dihedrals

    @property
    def impropers(self):
        """Get the list of improper angles in this :py:class:`Topology`."""
        count = self.impropers_count()
        impropers = np.zeros((count, 4), np.uint64)
        self.ffi.chfl_topology_impropers(self, impropers, c_uint64(count))
        return impropers

    def add_bond(self, i, j, order=None):
        """
        Add a bond between the atoms at indexes ``i`` and ``j`` in this
        :py:class:`Topology`, optionally setting the bond ``order``.
        """
        if order is None:
            self.ffi.chfl_topology_add_bond(self, c_uint64(i), c_uint64(j))
        else:
            self.ffi.chfl_topology_bond_with_order(
                self, c_uint64(i), c_uint64(j), chfl_bond_order(order)
            )

    def remove_bond(self, i, j):
        """
        Remove any existing bond between the atoms at indexes ``i`` and ``j``
        in this :py:class:`Topology`.

        This function does nothing if there is no bond between ``i`` and ``j``.
        """
        self.ffi.chfl_topology_remove_bond(self, c_uint64(i), c_uint64(j))



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from ctypes import c_double, c_uint64, c_char_p

from ._utils import CxxPointer, _call_with_growing_buffer, string_type
from .misc import ChemfilesError
from .property import Property


class Atom(CxxPointer):
    """
    An :py:class:`Atom` is a particle in the current :py:class:`Frame`. It
    stores the following atomic properties:

    - atom name;
    - atom type;
    - atom mass;
    - atom charge.

    The atom name is usually an unique identifier (``"H1"``, ``"C_a"``) while
    the atom type will be shared between all particles of the same type:
    ``"H"``, ``"Ow"``, ``"CH3"``.
    """

    def __init__(self, name, type=None):
        """
        Create a new :py:class:`Atom` with the given ``name``. If ``type`` is
        present, use it as the atom type. Else the atom name is used as atom
        type.
        """
        ptr = self.ffi.chfl_atom(name.encode("utf8"))
        super(Atom, self).__init__(ptr, is_const=False)
        if type:
            self.type = type

    def __copy__(self):
        return Atom.from_ptr(self.ffi.chfl_atom_copy(self))

    def __repr__(self):
        name = self.name
        type = self.type
        if type == name:
            return "Atom('{}')".format(name)
        else:
            return "Atom('{}', '{}')".format(name, type)

    @property
    def mass(self):
        """Get this :py:class:`Atom` mass, in atomic mass units."""
        mass = c_double()
        self.ffi.chfl_atom_mass(self, mass)
        return mass.value

    @mass.setter
    def mass(self, mass):
        """Set this :py:class:`Atom` mass, in atomic mass units."""
        self.ffi.chfl_atom_set_mass(self, c_double(mass))

    @property
    def charge(self):
        """
        Get this :py:class:`Atom` charge, in number of the electron charge *e*.
        """
        charge = c_double()
        self.ffi.chfl_atom_charge(self, charge)
        return charge.value

    @charge.setter
    def charge(self, charge):
        """
        Set this :py:class:`Atom` charge, in number of the electron charge *e*.
        """
        self.ffi.chfl_atom_set_charge(self, c_double(charge))

    @property
    def name(self):
        """Get this :py:class:`Atom` name."""
        return _call_with_growing_buffer(
            lambda buffer, size: self.ffi.chfl_atom_name(self, buffer, size), initial=32
        )

    @name.setter
    def name(self, name):
        """Set this :py:class:`Atom` name to ``name``."""
        self.ffi.chfl_atom_set_name(self, name.encode("utf8"))

    @property
    def type(self):
        """Get this :py:class:`Atom` type."""
        return _call_with_growing_buffer(
            lambda buffer, size: self.ffi.chfl_atom_type(self, buffer, size), initial=32
        )

    @type.setter
    def type(self, type):
        """Set this :py:class:`Atom` type to ``type``."""
        self.ffi.chfl_atom_set_type(self, type.encode("utf8"))

    @property
    def full_name(self):
        """
        Try to get the full name of this :py:class:`Atom` from its type. For
        example, the full name of "He" is "Helium". If the name can not be
        found, returns the empty string.
        """
        return _call_with_growing_buffer(
            lambda buff, size: self.ffi.chfl_atom_full_name(self, buff, size),
            initial=64,
        )

    @property
    def vdw_radius(self):
        """
        Try to get the Van der Waals radius of this :py:class:`Atom` from its
        type. If the radius can not be found, returns 0.
        """
        radius = c_double()
        self.ffi.chfl_atom_vdw_radius(self, radius)
        return radius.value

    @property
    def covalent_radius(self):
        """
        Try to get the covalent radius of this :py:class:`Atom` from its type.
        If the radius can not be found, returns 0.
        """
        radius = c_double()
        self.ffi.chfl_atom_covalent_radius(self, radius)
        return radius.value

    @property
    def atomic_number(self):
        """
        Try to get the atomic number of this :py:class:`Atom` from its type. If
        the atomic number can not be found, returns 0.
        """
        number = c_uint64()
        self.ffi.chfl_atom_atomic_number(self, number)
        return number.value

    def __iter__(self):
        # Disable automatic iteration from __getitem__
        raise TypeError("can not iterate over an atom")

    def __getitem__(self, name):
        """
        Get a property of this atom with the given ``name``, or raise an error
        if the property does not exists.
        """
        if not isinstance(name, string_type):
            raise ChemfilesError(
                "Invalid type {} for an atomic property name".format(type(name))
            )
        ptr = self.ffi.chfl_atom_get_property(self, name.encode("utf8"))
        return Property.from_ptr(ptr).get()

    def __setitem__(self, name, value):
        """
        Set a property of this atom, with the given ``name`` and ``value``.
        The new value overwrite any pre-existing property with the same name.
        """
        if not isinstance(name, string_type):
            raise ChemfilesError(
                "Invalid type {} for an atomic property name".format(type(name))
            )
        self.ffi.chfl_atom_set_property(self, name.encode("utf8"), Property(value))

    def properties_count(self):
        """Get the number of properties in this atom."""
        count = c_uint64()
        self.ffi.chfl_atom_properties_count(self, count)
        return count.value

    def list_properties(self):
        """Get the name of all properties in this atom."""
        count = self.properties_count()
        StringArray = c_char_p * count
        names = StringArray()
        self.ffi.chfl_atom_list_properties(self, names, count)
        return list(map(lambda n: n.decode("utf8"), names))



#!/usr/bin/env python
# -*- coding: utf8 -*-
"""
This script check that only whitelisted headers are included (transitivly) by
including chemfiles.h or chemfiles.hpp.
"""
import os
import sys
import re

ROOT = os.path.join(os.path.dirname(__file__), "..", "..")
GENERATED_HEADERS = ["chemfiles/config.hpp", "chemfiles/exports.hpp"]
ERRORS = 0

WHITELIST = [
    # standard C99 headers
    "stdbool.h", "stdint.h",
    # standard C++11 headers
    "iterator", "functional", "cstdint", "array", "utility", "cassert",
    "string", "memory", "exception", "limits", "algorithm", "stdexcept",
    "vector", "cmath", "type_traits", "unordered_map", "cfloat",
    # external headers
    "chemfiles/external/span.hpp",
    "chemfiles/external/optional.hpp",
    # chemfiles helper headers
    "chemfiles/exports.hpp",
    "chemfiles/config.hpp",
    "chemfiles/sorted_set.hpp",
    "chemfiles/unreachable.hpp",
    # chemfiles main headers
    "chemfiles/misc.hpp",
    "chemfiles/types.hpp",
    "chemfiles/Atom.hpp",
    "chemfiles/Frame.hpp",
    "chemfiles/Error.hpp",
    "chemfiles/Residue.hpp",
    "chemfiles/Property.hpp",
    "chemfiles/Topology.hpp",
    "chemfiles/UnitCell.hpp",
    "chemfiles/Trajectory.hpp",
    "chemfiles/Selection.hpp",
    "chemfiles/Connectivity.hpp",
    # chemfiles capi headers
    "chemfiles/capi/atom.h",
    "chemfiles/capi/selection.h",
    "chemfiles/capi/trajectory.h",
    "chemfiles/capi/residue.h",
    "chemfiles/capi/property.h",
    "chemfiles/capi/cell.h",
    "chemfiles/capi/frame.h",
    "chemfiles/capi/types.h",
    "chemfiles/capi/topology.h",
    "chemfiles/capi/misc.h",
]


def error(message):
    global ERRORS
    ERRORS += 1
    print(message)


def included_headers(path):
    includes = set()
    with open(path) as fd:
        for line in fd:
            if "#include" in line:
                matched = re.match("#include\s*[\"<](.*)[\">]", line)
                if not matched:
                    error("bad include in {}: {}".format(path, line))
                header = matched.groups()[0]
                includes.add(header)
                if header.startswith("chemfiles"):
                    if header not in GENERATED_HEADERS:
                        path = os.path.join(ROOT, "include", header)
                        includes.update(included_headers(path))
    return includes


def check_allowded(headers):
    for header in headers:
        if header not in WHITELIST:
            error("private header {} is publicly reachable".format(header))


if __name__ == '__main__':
    headers = included_headers(os.path.join(ROOT, "include", "chemfiles.h"))
    check_allowded(headers)

    headers = included_headers(os.path.join(ROOT, "include", "chemfiles.hpp"))
    check_allowded(headers)

    if ERRORS != 0:
        sys.exit(1)



# -* coding: utf-8 -*
import os
import sys
from ctypes import cdll, c_double, POINTER

try:
    from .external import EXTERNAL_CHEMFILES
except ImportError:
    EXTERNAL_CHEMFILES = False


class FindChemfilesLibrary(object):
    def __init__(self):
        self._cache = None

    def __call__(self):
        if self._cache is None:
            path = _lib_path()
            self._cache = cdll.LoadLibrary(path)

            from .ffi import set_interface
            from .ffi import CHFL_FRAME, CHFL_ATOM, chfl_vector3d

            set_interface(self._cache)
            # We update the arguments here, as ctypes can not pass a NULL value
            # as the last parameter
            self._cache.chfl_frame_add_atom.argtypes = [
                POINTER(CHFL_FRAME),
                POINTER(CHFL_ATOM),
                chfl_vector3d,
                POINTER(c_double),
            ]

            from .misc import _set_default_warning_callback

            _set_default_warning_callback()
        return self._cache


def _lib_path():
    if EXTERNAL_CHEMFILES:
        return EXTERNAL_CHEMFILES
    root = os.path.dirname(__file__)
    if sys.platform.startswith("darwin"):
        return os.path.join(root, "lib", "libchemfiles.dylib")
    elif sys.platform.startswith("linux"):
        return os.path.join(root, "lib", "libchemfiles.so")
    elif sys.platform.startswith("win"):
        candidates = [
            os.path.join(root, "bin", "libchemfiles.dll"),  # MinGW
            os.path.join(root, "bin", "chemfiles.dll"),  # MSVC
        ]
        for path in candidates:
            if os.path.isfile(path):
                _check_dll(path)
                return path
        raise ImportError("Could not find chemfiles DLL")
    else:
        raise ImportError("Unknown platform. Please edit this file")


def _check_dll(path):
    import struct
    import platform

    IMAGE_FILE_MACHINE_I386 = 332
    IMAGE_FILE_MACHINE_AMD64 = 34404

    machine = None
    with open(path, "rb") as fd:
        header = fd.read(2).decode(encoding="utf-8", errors="strict")
        if header != "MZ":
            raise ImportError(path + " is not a DLL")
        else:
            fd.seek(60)
            header = fd.read(4)
            header_offset = struct.unpack("<L", header)[0]
            fd.seek(header_offset + 4)
            header = fd.read(2)
            machine = struct.unpack("<H", header)[0]

    arch = platform.architecture()[0]
    if arch == "32bit":
        if machine != IMAGE_FILE_MACHINE_I386:
            raise ImportError("Python is 32-bit, but chemfiles.dll is not")
    elif arch == "64bit":
        if machine != IMAGE_FILE_MACHINE_AMD64:
            raise ImportError("Python is 64-bit, but chemfiles.dll is not")
    else:
        raise ImportError("Could not determine pointer size of Python")


_get_c_library = FindChemfilesLibrary()



#!/usr/bin/env python
# -*- coding: utf8 -*-
"""
This script check for some functions we don't want to use because they are
better wrappers in chemfiles
"""
import os
import sys
import re
from glob import glob

ROOT = os.path.join(os.path.dirname(__file__), "..", "..")
ERRORS = 0

PARSING_FUNCTIONS = {
    # Match 'std::<...>',  '::<...>' and '<...>'
    r"(?P<name>\b(std)?(::)?stof\b)": "parse<double>",
    r"(?P<name>\b(std)?(::)?stod\b)": "parse<double>",
    r"(?P<name>\b(std)?(::)?stold\b)": "parse<double>",
    r"(?P<name>\b(std)?(::)?stoi\b)": "parse<long long>",
    r"(?P<name>\b(std)?(::)?stol\b)": "parse<long long>",
    r"(?P<name>\b(std)?(::)?stoll\b)": "parse<long long>",
    r"(?P<name>\b(std)?(::)?stoul\b)": "parse<size_t>",
    r"(?P<name>\b(std)?(::)?stoull\b)": "parse<size_t>",
}

CAPI_FUNCTIONS = {
    r"(?P<name>\bstatic_cast<size_t>\b)": "checked_cast",
    r"(?P<name>\bnew\b)": "shared_allocator::make_shared",
    r"(?P<name>\bdelete\b)": "shared_allocator::free",
    r"(?P<name>\bmalloc\b)": "shared_allocator::make_shared",
    r"(?P<name>\bfree\b)": "shared_allocator::free",
}


def error(message):
    global ERRORS
    ERRORS += 1
    print(message)


def check_code(path, replacements):
    with open(path) as fd:
        for i, line in enumerate(fd):
            for bad, replacement in replacements.items():
                match = re.search(bad, line)
                if match and "NOLINT" not in line:
                    error(
                        "Please replace '{}' by '{}' at {}:{}".format(
                            match.group("name"),
                            replacement,
                            os.path.relpath(path),
                            i + 1,
                        )
                    )


if __name__ == "__main__":
    for file in glob(os.path.join(ROOT, "include", "chemfiles", "*.hpp")):
        if file.endswith("utils.hpp"):
            continue
        check_code(file, PARSING_FUNCTIONS)

    for file in glob(os.path.join(ROOT, "include", "chemfiles", "*", "*.hpp")):
        check_code(file, PARSING_FUNCTIONS)

    for file in glob(os.path.join(ROOT, "src", "*.cpp")):
        check_code(file, PARSING_FUNCTIONS)

    for file in glob(os.path.join(ROOT, "src", "*", "*.cpp")):
        check_code(file, PARSING_FUNCTIONS)

    for file in glob(os.path.join(ROOT, "src", "capi", "*.cpp")):
        check_code(file, CAPI_FUNCTIONS)

    if ERRORS != 0:
        sys.exit(1)



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from ctypes import c_uint64
import numpy as np

from ._utils import CxxPointer, _call_with_growing_buffer
from .ffi import chfl_match


class Selection(CxxPointer):
    """
    Select atoms in a :py:class:`Frame` with a selection language.

    The selection language is built by combining basic operations. Each basic
    operation follows the ``<selector>[(<variable>)] <operator> <value>``
    structure, where ``<operator>`` is a comparison operator in
    ``== != < <= > >=``. Refer to the `full documentation
    <selections-doc>`_ to know the allowed selectors and how to use them.

    .. selections-doc: http://chemfiles.rtfd.io/en/latest/selections.html
    """

    def __init__(self, selection):
        """
        Create a new :py:class:`Selection` from the given ``selection`` string.
        """
        ptr = self.ffi.chfl_selection(selection.encode("utf8"))
        super(Selection, self).__init__(ptr, is_const=False)

    def __copy__(self):
        return Selection.from_ptr(self.ffi.chfl_selection_copy(self))

    def __repr__(self):
        return "Selection('{}')".format(self.string)

    @property
    def size(self):
        """
        Get the size of this :py:class:`Selection`.

        The size of a selection is the number of atoms we are selecting
        together. This value is 1 for the 'atom' context, 2 for the 'pair' and
        'bond' context, 3 for the 'three' and 'angles' contextes and 4 for the
        'four' and 'dihedral' contextes.
        """
        size = c_uint64()
        self.ffi.chfl_selection_size(self, size)
        return size.value

    @property
    def string(self):
        """
        Get the selection string used to create this :py:class:`Selection`.
        """
        return _call_with_growing_buffer(
            lambda buff, n: self.ffi.chfl_selection_string(self, buff, n), initial=128
        )

    def evaluate(self, frame):
        """
        Evaluate a :py:class:`Selection` for a given :py:class:`Frame`, and
        return a list of matching atoms, either as a list of index or a list
        of tuples of indexes.
        """
        matching = c_uint64()
        self.ffi.chfl_selection_evaluate(self, frame, matching)

        matches = np.zeros(matching.value, chfl_match)
        self.ffi.chfl_selection_matches(self, matches, matching)

        size = self.size
        result = []
        for match in matches:
            assert match[0] == size
            atoms = match[1]
            if size == 1:
                result.append(atoms[0])
            elif size == 2:
                result.append((atoms[0], atoms[1]))
            elif size == 3:
                result.append((atoms[0], atoms[1], atoms[2]))
            elif size == 4:
                result.append((atoms[0], atoms[1], atoms[2], atoms[3]))
        return result



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from ctypes import c_bool, c_uint64, c_char_p
import numpy as np

from ._utils import CxxPointer, _call_with_growing_buffer, string_type
from .misc import ChemfilesError
from .property import Property


class ResidueAtoms(object):
    """Proxy object to get the atomic indexes in a residue"""

    def __init__(self, residue):
        self.residue = residue
        # Use a cache for indexes
        self.indexes = None

    def __len__(self):
        """Get the current number of atoms in this :py:class:`Residue`."""
        if self.indexes is not None:
            return len(self.indexes)
        else:
            count = c_uint64()
            self.residue.ffi.chfl_residue_atoms_count(self.residue, count)
            return count.value

    def __contains__(self, index):
        """
        Check if the :py:class:`Residue` contains the atom at index ``atom``.
        """
        if self.indexes is not None:
            return index in self.indexes
        else:
            result = c_bool()
            self.residue.ffi.chfl_residue_contains(self.residue, c_uint64(index), result)
            return result.value

    def __getitem__(self, i):
        """
        Get the atomic index number ``i`` in the associated :py:class:`Residue`.
        """
        if self.indexes is None:
            count = len(self)
            self.indexes = np.zeros(count, np.uint64)
            self.residue.ffi.chfl_residue_atoms(self.residue, self.indexes, c_uint64(count))

        return self.indexes[i]

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __repr__(self):
        return "[" + ", ".join([i.__repr__() for i in self]) + "]"

    def append(self, atom):
        """Add the atom index ``atom`` in the :py:class:`Residue`."""
        self.residue.ffi.chfl_residue_add_atom(self.residue, c_uint64(atom))
        # reset the cache for indexes
        self.indexes = None


class Residue(CxxPointer):
    """
    A :py:class:`Residue` is a group of atoms belonging to the same logical
    unit. They can be small molecules, amino-acids in a protein, monomers in
    polymers, etc.
    """

    def __init__(self, name, resid=None):
        """
        Create a new :py:class:`Residue` from a ``name`` and optionally a
        residue id ``resid``.
        """

        if resid:
            ptr = self.ffi.chfl_residue_with_id(name.encode("utf8"), c_uint64(resid))
        else:
            ptr = self.ffi.chfl_residue(name.encode("utf8"))
        super(Residue, self).__init__(ptr, is_const=False)

    def __copy__(self):
        return Residue.from_ptr(self.ffi.chfl_residue_copy(self))

    def __repr__(self):
        return "Residue('{}') with {} atoms".format(self.name, len(self.atoms))

    @property
    def name(self):
        """Get the name of this :py:class:`Residue`."""
        return _call_with_growing_buffer(
            lambda buff, size: self.ffi.chfl_residue_name(self, buff, size), initial=32
        )

    @property
    def id(self):
        """Get the :py:class:`Residue` index in the initial topology."""
        id = c_uint64()
        self.ffi.chfl_residue_id(self, id)
        return id.value

    @property
    def atoms(self):
        return ResidueAtoms(self)

    def __iter__(self):
        # Disable automatic iteration from __getitem__
        raise TypeError("can not iterate over a residue")

    def __getitem__(self, name):
        """
        Get a property of this residude with the given ``name``, or raise an
        error if the property does not exists.
        """
        if not isinstance(name, string_type):
            raise ChemfilesError(
                "Invalid type {} for a residue property name".format(type(name))
            )
        ptr = self.ffi.chfl_residue_get_property(self, name.encode("utf8"))
        return Property.from_ptr(ptr).get()

    def __setitem__(self, name, value):
        """
        Set a property of this residue, with the given ``name`` and ``value``.
        The new value overwrite any pre-existing property with the same name.
        """
        if not isinstance(name, string_type):
            raise ChemfilesError(
                "Invalid type {} for a residue property name".format(type(name))
            )
        self.ffi.chfl_residue_set_property(self, name.encode("utf8"), Property(value))

    def properties_count(self):
        """Get the number of properties in this residue."""
        count = c_uint64()
        self.ffi.chfl_residue_properties_count(self, count)
        return count.value

    def list_properties(self):
        """Get the name of all properties in this residue."""
        count = self.properties_count()
        StringArray = c_char_p * count
        names = StringArray()
        self.ffi.chfl_residue_list_properties(self, names, count)
        return list(map(lambda n: n.decode("utf8"), names))



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import sys
from ctypes import create_string_buffer, c_uint64

from .clib import _get_c_library
from .misc import ChemfilesError, _last_error


if sys.version_info >= (3, 0):
    string_type = str
else:
    string_type = basestring


class CxxPointer(object):
    __frozen = False

    def __init__(self, ptr, is_const=True):
        _check_handle(ptr)
        self.__ptr = ptr
        self.__is_const = is_const
        self._as_parameter_ = self.__ptr
        self.__frozen = True

    def __del__(self):
        """Free the memory associated with this instance"""
        if hasattr(self, "__ptr"):
            self.ffi.chfl_free(self)

    def __setattr__(self, key, value):
        if self.__frozen and not hasattr(self, key):
            raise TypeError(
                "Can not add new attributes to this {}".format(self.__class__.__name__)
            )
        object.__setattr__(self, key, value)

    @classmethod
    def from_ptr(cls, ptr):
        """Create a new instance from a mutable pointer"""
        new = cls.__new__(cls)
        super(cls, new).__init__(ptr, is_const=False)
        return new

    @classmethod
    def from_const_ptr(cls, ptr):
        """Create a new instance from a const pointer"""
        new = cls.__new__(cls)
        super(cls, new).__init__(ptr, is_const=True)
        return new

    @property
    def ptr(self):
        """Get the **mutable** C++ pointer for this object"""
        if self.__is_const:
            raise ChemfilesError("Trying to use a const pointer for mutable access")
        else:
            return self.__ptr

    @property
    def const_ptr(self):
        """Get the **const** C++ pointer for this object"""
        return self.__ptr

    @property
    def ffi(self):
        """Allow to access the C interface from any instance of CxxPointer"""
        return _get_c_library()

    @classmethod
    def from_param(cls, parameter):
        if parameter is None:
            raise TypeError("Can not pass None as a " + cls.__name__)
        if not isinstance(parameter, cls):
            raise TypeError(
                "Can not pass " + parameter.__class__.__name__ + " as a " + cls.__name__
            )
        return parameter


def _check_return_code(status, _function, _arguments):
    """Check that the function call was OK, and raise an exception if needed"""
    if status.value != 0:
        raise ChemfilesError(_last_error())


def _check_handle(handle):
    """Check that C allocated pointers are not NULL"""
    try:
        handle.contents
    except ValueError:
        raise ChemfilesError(_last_error())


def _call_with_growing_buffer(function, initial):
    """
    Call ``function`` with a growing buffer until the buffer is big enough.
    Use ``initial`` as the initial buffer size.
    """

    def buffer_was_big_enough(buffer):
        if len(buffer) < 2:
            return False
        else:
            return buffer[-2] == b"\0"

    size = initial
    buffer = create_string_buffer(b"\0", size)
    function(buffer, c_uint64(size))

    while not buffer_was_big_enough(buffer):
        # Grow the buffer and retry
        size *= 2
        buffer = create_string_buffer(b"\0", size)
        function(buffer, c_uint64(size))

    return buffer.value.decode("utf8")



# -*- coding=utf-8 -*-
import chemfiles
import sys
from wheel.bdist_wheel import bdist_wheel
from skbuild import setup


class universal_wheel(bdist_wheel):
    # Workaround until https://github.com/pypa/wheel/issues/185 is resolved
    def get_tag(self):
        tag = bdist_wheel.get_tag(self)
        return ("py2.py3", "none") + tag[2:]


with open("requirements.txt", "r") as fd:
    requirements = list(filter(bool, (line.strip() for line in fd)))
    if sys.hexversion >= 0x03040000:
        requirements = list(filter(lambda r: r != "enum34", requirements))


long_description = ""
with open("README.md", "r") as fd:
    for line in fd:
        # remove github badges
        if not line.startswith("[!["):
            long_description += line

setup(
    name="chemfiles",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=chemfiles.__version__,
    author="Guillaume Fraux",
    author_email="luthaf@luthaf.fr",
    description="Read and write chemistry trajectory files",
    keywords="chemistry computational cheminformatics files formats",
    url="http://github.com/chemfiles/chemfiles.py",
    packages=["chemfiles"],
    zip_safe=False,
    install_requires=requirements,
    setup_requires=["scikit-build"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    cmake_install_dir="chemfiles",
    cmake_args=['-DCMAKE_OSX_DEPLOYMENT_TARGET:STRING=10.9'],
    cmdclass={"bdist_wheel": universal_wheel},
)



#!/usr/bin/env python
# -*- coding: utf8 -*-
"""
A small script checking that all the C API functions are documented, and have
an example.
"""
import os
import sys

ROOT = os.path.join(os.path.dirname(__file__), "..", "..")
ERRORS = 0


def error(message):
    global ERRORS
    ERRORS += 1
    print(message)


def documented_functions():
    functions = []
    DOCS = os.path.join(ROOT, "doc", "src", "capi")
    for (root, _, paths) in os.walk(DOCS):
        for path in paths:
            with open(os.path.join(root, path)) as fd:
                for line in fd:
                    if line.startswith(".. doxygenfunction::"):
                        name = line.split()[2]
                        functions.append(name)
    return functions


def functions_in_outline():
    functions = ["chfl_last_error", "chfl_clear_errors", "chfl_version"]
    DOCS = os.path.join(ROOT, "doc", "src", "capi")
    for (root, _, paths) in os.walk(DOCS):
        for path in paths:
            with open(os.path.join(root, path)) as fd:
                for line in fd:
                    if ":cpp:func:" in line:
                        name = line.split("`")[1]
                        functions.append(name)
    return functions


def function_name(line):
    assert(line.startswith("CHFL_EXPORT"))
    splitted = line.split()
    if splitted[2].startswith("chfl_"):
        name = splitted[2].split('(')[0]
    elif splitted[3].startswith("chfl_"):
        name = splitted[3].split('(')[0]
    else:
        raise RuntimeError("Could not get function name in '" + line + "'")
    return name


def all_functions():
    functions = []
    HEADERS = os.path.join(ROOT, "include", "chemfiles", "capi")
    for (root, _, paths) in os.walk(HEADERS):
        for path in paths:
            with open(os.path.join(root, path)) as fd:
                for line in fd:
                    if line.startswith("CHFL_EXPORT"):
                        functions.append(function_name(line))
    return functions


def check_examples():
    HEADERS = os.path.join(ROOT, "include", "chemfiles", "capi")
    for (root, _, paths) in os.walk(HEADERS):
        for path in paths:
            with open(os.path.join(root, path)) as fd:
                in_doc = False
                example_found = False

                for line in fd:
                    if "@example" in line and in_doc:
                        example_found = True
                        path = line.split('{')[1].split('}')[0]
                        if not os.path.exists(os.path.join(ROOT, "tests", "doc", path)):
                            error("Missing example file at {}".format(path))

                    if line.startswith("///"):
                        in_doc = True
                    elif line.startswith("CHFL_EXPORT"):
                        in_doc = False
                        if not example_found:
                            name = function_name(line)
                            error("Missing example for {}".format(name))
                        example_found = False


if __name__ == '__main__':
    docs = documented_functions()
    outline = functions_in_outline()
    for function in all_functions():
        if function not in docs:
            error("Missing documentation for {}".format(function))
        if function not in outline:
            error("Missing outline for {}".format(function))
    check_examples()

    # C and fortran standard only allow extern names up to 31 characters
    for function in all_functions():
        if len(function) > 31:
            error("Function name {} is too long".format(function))

    if ERRORS != 0:
        sys.exit(1)



#!/usr/bin/env python
# -* coding: utf-8 -*

"""
This script reads the standard connectivty table provided by
the PDB and produces a C++ header file that maps residue names
to their connectivity by atom name.

WARNING: The entire PDB dictionary is quite large, so it is recomended to
only use part of it (see usage).

The online version of this table is available here:

http://ftp.wwpdb.org/pub/pdb/data/monomers/het_dictionary.txt

"""

import sys
import re

class Interner:
    def __init__(self):
        self.list = []

    def get(self, string):
        for (i, known) in enumerate(self.list):
            if string == known:
                return i
        self.list.append(string)
        return len(self.list) - 1


INTERNER = Interner()


def join_and_wrap_80(indent, list):
    result = ""
    line = indent
    for string in list:
        line += string + ", "
        if len(line) > 72:
            result += line + "\n"
            line = indent
    if line != indent:
        result += line
    return result


class Residue:
    def __init__(self, code):
        self.code = code.strip()
        self.atoms = {}

    def __str__(self):
        connectivity = []
        for atom, connections in self.atoms.iteritems():
            for connected in connections:
                connectivity.append('{{{}, {}}}'.format(atom, connected))
        connectivity = join_and_wrap_80("        ", connectivity)
        return '{{"{}", {{\n{}    }}}}'.format(self.code, connectivity)

    def add_atom(self, name, connected):
        self.atoms[INTERNER.get(name)] = [INTERNER.get(s) for s in connected]

def read_residues(path, accepted_residues=None):
    residues = []
    read_residue = False

    with open(path) as f:
        for line in f:
            line = line.replace(r'"', r'\"')
            if line[:7] == "RESIDUE":
                read_residue = False
                if accepted_residues is None or line[10:13] in accepted_residues:
                    residues.append(Residue(line[10:13]))
                    read_residue = True
            if line[:6] == "CONECT" and read_residue:
                splitted = re.split(r'\s+', line)[1:-1]
                residues[-1].add_atom(splitted[0], splitted[2:])

    return residues

HEADER = """// Chemfiles, a modern library for chemistry file reading and writing
// Copyright (C) Guillaume Fraux and contributors -- BSD license

// !!!! AUTO-GENERATED FILE !!!! Do not edit. See pdb_connectivity.py for the code.
// The data comes from the PDB's connectivity table:
// http://ftp.wwpdb.org/pub/pdb/data/monomers/het_dictionary.txt

#include "chemfiles/pdb_connectivity.hpp"
using namespace chemfiles;

"""

def write_elements(path, residues):
    with open(path, "w") as fd:
        fd.write(HEADER)

        fd.write("// generated for the following residues\n")
        fd.write(join_and_wrap_80("// ", [r.code for r in residues]))
        fd.write("\n\n")

        fd.write("const std::vector<std::string> InternedName::INTERNER_ = {\n")
        fd.write(join_and_wrap_80("    ", ['"{}"'.format(name) for name in INTERNER.list]))
        fd.write("};\n\n\n")

        fd.write("const PDBConnectivity::PDBConnectMap PDBConnectivity::PDB_CONNECTIVITY_MAP_ = {\n")
        for residue in residues:
            fd.write("    " + str(residue) + ",\n")
        fd.write("};\n\n")


def usage():
    print(sys.argv[0] + " path/to/het_dictionary.txt pdb_connectivity.cpp [Residue names...]")


if __name__ == "__main__":
    if len(sys.argv) < 3:
        usage()
        sys.exit(0)

    if len(sys.argv) >= 4:
        residue_names = set(sys.argv[3:])
        residue_names.update({
            "GLU", "GLY", "PHE", "LEU", "SER", "LYS", "CYS", "TRP", "TYR",
            "PRO", "HIS", "GLN", "ARG", "ILE", "MET", "THR", "ASN", "ASP",
            "VAL", "ALA", "A  ", "U  ", "G  ", "C  ", "DA ", "DT ", "DG ", "DC "
        });
        residues = read_residues(sys.argv[1], residue_names)
    else:
        residues = read_residues(sys.argv[1])

    write_elements(sys.argv[2], residues)



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import ctypes
from ctypes import c_uint64, c_char_p

from ._utils import CxxPointer
from .frame import Frame, Topology
from .misc import ChemfilesError


class Trajectory(CxxPointer):
    """
    A :py:class:`Trajectory` is a chemistry file on the hard drive. It is the
    main entry point of Chemfiles.
    """

    def __init__(self, path, mode="r", format=""):
        """
        Open the :py:class:`Trajectory` at the given ``path`` using the
        given ``mode`` and optional specific file ``format``.

        Valid modes are ``'r'`` for read, ``'w'`` for write and ``'a'`` for
        append.

        The specific file format is needed when the file format does not match
        the extension, or when there is not standard extension for this format.
        If `format` is an empty string, the format will be guessed from the
        file extension.
        """
        self.__closed = False
        # Store mode and format for __repr__
        self.__mode = mode
        self.__format = format
        ptr = self.ffi.chfl_trajectory_with_format(
            path.encode("utf8"), mode.encode("utf8"), format.encode("utf8")
        )
        super(Trajectory, self).__init__(ptr, is_const=False)

    def __check_opened(self):
        if self.__closed:
            raise ChemfilesError("Can not use a closed Trajectory")

    def __del__(self):
        if not self.__closed and hasattr(self, "__ptr"):
            self.close()

    def __enter__(self):
        self.__check_opened()
        return self

    def __exit__(self, *args):
        self.close()

    def __iter__(self):
        self.__check_opened()
        for step in range(self.nsteps):
            yield self.read_step(step)

    def __repr__(self):
        return "Trajectory('{}', '{}', '{}')".format(self.path, self.__mode, self.__format)

    def read(self):
        """
        Read the next step of the :py:class:`Trajectory` and return the
        corresponding :py:class:`Frame`.
        """
        self.__check_opened()
        frame = Frame()
        self.ffi.chfl_trajectory_read(self, frame)
        return frame

    def read_step(self, step):
        """
        Read a specific ``step`` in the :py:class:`Trajectory` and return the
        corresponding :py:class:`Frame`.
        """
        self.__check_opened()
        frame = Frame()
        self.ffi.chfl_trajectory_read_step(self, c_uint64(step), frame)
        return frame

    def write(self, frame):
        """Write a :py:class:`Frame` to the :py:class:`Trajectory`."""
        self.__check_opened()
        self.ffi.chfl_trajectory_write(self, frame)

    def set_topology(self, topology, format=""):
        """
        Set the :py:class:`Topology` associated with a :py:class:`Trajectory`.
        This :py:class:`Topology` will be used when reading and writing the
        files, replacing any :py:class:`Topology` in the frames or files.

        If ``topology`` is a :py:class:`Topology` instance, it is used
        directly. If ``topology`` is a string, the first :py:class:`Frame` of
        the corresponding file is read, and the topology of this frame is used.

        When reading from a file, if ``format`` is not the empty string, the
        code uses this file format instead of guessing it from the file
        extension.
        """
        self.__check_opened()
        if isinstance(topology, Topology):
            self.ffi.chfl_trajectory_set_topology(self, topology)
        else:
            self.ffi.chfl_trajectory_topology_file(
                self, topology.encode("utf8"), format.encode("utf8")
            )

    def set_cell(self, cell):
        """
        Set the :py:class:`UnitCell` associated with a :py:class:`Trajectory`.
        This :py:class:`UnitCell` will be used when reading and writing the
        files, replacing any :py:class:`UnitCell` in the frames or files.
        """
        self.__check_opened()
        self.ffi.chfl_trajectory_set_cell(self, cell)

    @property
    def nsteps(self):
        """
        Get the number of steps (the number of frames) in a
        :py:class:`Trajectory`.
        """
        self.__check_opened()
        nsteps = c_uint64()
        self.ffi.chfl_trajectory_nsteps(self, nsteps)
        return nsteps.value

    @property
    def path(self):
        """Get the path used to open this  :py:class:`Trajectory`."""
        self.__check_opened()
        path = c_char_p()
        self.ffi.chfl_trajectory_path(self, path)
        return path.value.decode('utf-8')

    def close(self):
        """
        Close the :py:class:`Trajectory` and write any buffered content to the
        hard drive.
        """
        self.__check_opened()
        self.__closed = True
        self.ffi.chfl_trajectory_close(self)
